/* 
   IHELP Simulation - PersAgent V1.0 

   Written by: 
      Alan Fedoruk                                           -- amf673@mail.usask.ca
      MADMUC Lab                                             -- http://bistrica.usask.ca/madmuc/index.htm
      Dept. of Computer Science, University of Saskatchewan. -- http://www.cs.usask.ca

   Copyright (c) 2001 Alan M. Fedoruk
*/

package ihelpnorep;

// Import the fipa-os classes
import fipaos.agent.*;
import fipaos.agent.conversation.*;
import fipaos.agent.task.*;
import fipaos.mts.*;
import fipaos.ont.fipa.*;
import fipaos.ont.fipa.fipaman.*;
import fipaos.parser.acl.ACLMessage;
import fipaos.platform.ams.AMSRegistrationException;
import fipaos.platform.df.DFRegistrationException;
import fipaos.util.DIAGNOSTICS;

// Import java classes
import java.util.*;
import java.io.*;
import ihelp.*;


/** A users Personal Agent. 
    Stores a little about a user and knows if the user is on line 
    or off.  */
public class PersAgent extends FIPAOSAgent 
{

   /* This Agent type */
   private static final String AGENT_TYPE = IHCONSTANTS.PERSAGENT;

   /* The AID of the UI agent associated with this Personal Agent */
   private AgentID _uiagent = null;

   /* The name of my user */
   private String _user_name;  

   /* A vector of the things my user knows about */
   private Vector _keywords; 

   /* The status of my user true = online, false = offline */
   private boolean _status; 

   /* A matchmakers agent */
   private AgentID _matchmaker;

   /* A random number generator for simulated failures */
   private Random _r = new Random();
  

   /** standard RepAgent constructor */
   public PersAgent( String platform, String name, String ownership) 
   {
   
      super( platform, name, ownership);
    
      // Instantiate a new IdleTask to receieve all incoming 
      // communications which are not associated with any-other Task

      setListenerTask( new IdleTask() );
  
      setUIAgent(null); 
      setUserName(null);  
      _keywords = new Vector(); 
      _keywords.addElement("java");
      setStatus(false); 

  // Register with the local AMS (i.e. the platform)
      try 
      {
         registerWithAMS();
         DIAGNOSTICS.println( "Registered with AMS", this, DIAGNOSTICS.LEVEL_MAX );
      } catch (AMSRegistrationException amsre) 
      {
         DIAGNOSTICS.println(amsre, this, DIAGNOSTICS.LEVEL_MAX );
         String reason = amsre.getExceptionReason();
      
         if ( reason == null || !reason.equals( FIPAMANCONSTANTS.AGENT_ALREADY_REGISTERED)) 
         {
            shutdown();
            return;
         }
      } // end catch 
    
      // Register with the local DF
      try 
      {
         registerWithDF( name );
         DIAGNOSTICS.println("Registered with DF", this, DIAGNOSTICS.LEVEL_MAX );
      } catch (DFRegistrationException dfre) 
      {
         DIAGNOSTICS.println(dfre, this, DIAGNOSTICS.LEVEL_MAX );
      
         String reason = dfre.getExceptionReason();
      
         if ( reason == null || !reason.equals( FIPAMANCONSTANTS.AGENT_ALREADY_REGISTERED) ) 
         {
            shutdown();
            return;
         }
      } // end catch

      _tm.newTask( new FindMMTask() );

   } // PersAgent()
  



   public void donePersAgent_FindMMTask(Task t)
   {
   }


   /** Invoked when this Agent is about to be shutdown - this overrides the
       shutdown() method in the FIPAOSAgent class to enable the CountAgentGUI
       associated with this CountAgent to be removed */
   public synchronized void shutdown() 
   {
      if (registeredWithDF() == true) 
      {
         // Attempt to deregister with DF
         try 
         {
            deregisterWithDF();
            DIAGNOSTICS.println("Deregistered with DF", this, DIAGNOSTICS.LEVEL_MAX );
         } catch (DFRegistrationException dfre) 
         {
            // A problem deregistering occured..
            String reason = dfre.getExceptionReason();
            DIAGNOSTICS.println(dfre + ":" + reason, this, DIAGNOSTICS.LEVEL_MAX ); 
         } // end catch
      } // if 
    
      if (registeredWithAMS() == true) 
      {
         // Attempt to deregister with AMS
         try 
         {
            deregisterWithAMS();
            DIAGNOSTICS.println("Deregistered with AMS", this, DIAGNOSTICS.LEVEL_MAX ); 

         } catch (AMSRegistrationException amsre) 
         {
            // A problem deregistering occured.
            String reason = amsre.getExceptionReason();
            DIAGNOSTICS.println(amsre + ":" + reason, this, DIAGNOSTICS.LEVEL_MAX );
         } // catch
      } // if 

      // Now call shutdown in the agent shell to release the core components
      super.shutdown();

   } // shutdown()


   /** set the user_name for the agent */
   public void setUserName(String name) 
   {
      _user_name = name; 
   } // setUserName()


   /** get the user_name for the agent */
   public String getUserName() 
   {
      return _user_name; 
   } // getUserName()


   /** set the AgentID of the user interface agent */
   public void setUIAgent(AgentID aid) 
   {
      _uiagent = aid; 
   } // setUIAgent()


   /** get the AgentID of the user interface agent */
   public AgentID getUIAgent() 
   {
      return (_uiagent); 
   } // getUIAgent()


   /** add the given keyword to the list of keywords */
   public void addKeyword(String kw) 
   {
      if (!hasKeyword(kw)) 
         _keywords.addElement(kw);
   } // addKeyword()


   /** get a string of the list of keywords */
   public String getKeywords() 
   {
      String res = "";
      for ( int i=0; i < _keywords.size(); i++) 
      {
         res = res + " " + (String)_keywords.elementAt(i);
      }
      return res;
   }


   /** return true if the keyword is in the list. */ 
   public boolean hasKeyword( String kw) 
   {
      return _keywords.contains(kw); 
   } // hasKeyword()


   /** set the status, true is user is available, false otherwise. */ 
   public void setStatus(boolean st) 
   {
      _status = st;
   } // setStatus()


   /** get the status. */
   public boolean getStatus() 
   {
      return(_status);
   } // setStatus()

   /** get the String status. */
   public String getStrStatus() 
   {
      if ( getStatus()) 
         return ( IHCONSTANTS.ONLINE);
      else 
         return ( IHCONSTANTS.OFFLINE);
   } // setStatus()


   /** generate a random number, and if the number is in the 
       range to fail, shutdown this agent.  */
   public void checkForFailure()
   {
      float r = _r.nextFloat();
      if ( r <= IHCONSTANTS.FAIL_RATE) 
      {
         System.out.println(getAID().getName() + " is failing ");
         shutdown();
      }
   }

   /** Send a reply message in response to the given ACL message
       @param acl The ACL object to reply to
       @param reason Any additional message
   */
   public boolean sendReply( ACL orig_acl, String performative, String content )
   {
      try
      {
         // Duplicate ACL
         ACL acl = (ACL) orig_acl.clone();
            
         // Setup message
         acl.setPerformative( performative );
         acl.setReceiverAID( acl.getSenderAID() );
         acl.setSenderAID( getAID() );
         acl.setInReplyTo( acl.getReplyWith() );
         acl.setReplyWith( null );
            
         // Set content
         acl.setContentObject( content );
            
         // Attempt to send the reply
         forward( acl );
            
         // We're done
         return true;
      }
      catch( Throwable t )
      {
         DIAGNOSTICS.println( "Error occured when sending reply", t, this,
         DIAGNOSTICS.LEVEL_MAX );
         return false;
      }
   } // sendReply()

   
   public boolean sendReply( ACL orig_acl, String performative)
   {
      return sendReply(orig_acl, performative, ""); 
   }


   /** Get the number of current conversations */
   public int getConvs()
   {
      return getCM().getConvCount();
   }


   /** remove the given conversation */
   public void removeConv(Conversation conv)
   {
      try 
      {
         getCM().removeConversation(conv.getConversationID());
      }
      catch (Exception e)
      {
         System.out.println(e);
      }
   }






   /** IdleTask 
   */
   public class IdleTask extends Task 
   {


      public IdleTask() 
      {
      } // IdleTask()
    
    
      protected void startTask() 
      {
      } // StartTask()


      public void handleRequest( Conversation conv ) 
      {
         // Get the type of request
         System.out.println( "PersAgent " + getAID().getName() + " Tasks: " + _tm.getTaskCount() + " Convs: " + getConvs());
         ACL acl = conv.getACL( conv.getLatestMessageIndex() );

         // if this is the active agent, process the request.
         if ( acl.getOntology().equalsIgnoreCase( IHCONSTANTS.IHELP_ONT ) ) 
         {
            newTask ( new HandleRequestTask(acl, conv), conv); 
         }
         else 
            sendNotUnderstood(acl);

      } // handleRequest()


      public void donePersAgent_HandleRequestTask(Task t)
      {
         // checkForFailure();
         // newTask( new StateTask());
      }
    
   } // IdleTask()



   public class FindMMTask extends Task 
   {

      public FindMMTask()
      {
      }

 
      protected void startTask() 
      {
	 // Search the DF for a MM
         DFAgentDescription df_desc = new DFAgentDescription();
         // Create a ServiceDescription of the match-maker agents
         ServiceDescription sd = new ServiceDescription();
         sd.setServiceName(IHCONSTANTS.MMAGENT);
         sd.setServiceType(IHCONSTANTS.MMAGENT); 		
         df_desc.setAgentServices( new ServiceDescription[] { sd } );
			
         // Start a search for other Agents
         searchDF( df_desc );
         DIAGNOSTICS.println( "PersAgent: Started DF search", DIAGNOSTICS.LEVEL_4 );
			

         // register with an MM 
         
      } // startTask()


      /** This method is invoked when a DFSearchTask has completed
          succesfully.
      */ 
      public void DFSearchResults( DFAgentDescription descs[] )
      {
         // Check we've got some results - if not just return
         if ( descs == null || descs.length <= 0 )
         {
            _matchmaker = null;
            DIAGNOSTICS.println("Did not find a matchmaker", DIAGNOSTICS.LEVEL_3);
         }
         else 
         {
            _matchmaker = descs[ 0 ].getAgentID();
            DIAGNOSTICS.println("Found a matchmaker: " + _matchmaker.getName(), this, DIAGNOSTICS.LEVEL_3);
            ACL reg = new ACL(); 
            reg.setOntology(IHCONSTANTS.IHELP_ONT);
            reg.setReceiverAID( _matchmaker); 
            reg.setSenderAID( getAID());
            reg.setPerformative( FIPACONSTANTS.REQUEST);
            reg.setContentObject( IHCONSTANTS.REGISTER + " " + getAID().toString());
            reg.setProtocol( FIPACONSTANTS.FIPA_REQUEST );
            forward(reg);
         }

      } // DFSearchResults()


      public void handleAgree(Conversation c)
      {
      }

      public void handleInform(Conversation c)
      { // once get a message back from matchmaker, we are done. 
         done();
      }

   }

   /** Handles all incoming i-help requests
   */
   public class HandleRequestTask extends Task
   {


      private ACL _acl;
      private Conversation _conv;
      

      public HandleRequestTask ( ACL acl, Conversation conv) 
      {
         _acl = acl;
         _conv = conv;
      }
      

      public void startTask() 
      {

         String s = (String)_acl.getContentObject();
         StringTokenizer st = new StringTokenizer(s);
         int n = st.countTokens();
         String function = st.nextToken(); 
         if (function.equals(IHCONSTANTS.KEYWORD_ADD))
         {
            if (n == 2) 
            {
               addKeyword(st.nextToken()); 
               sendReply(_acl, FIPACONSTANTS.AGREE);
               sendReply(_acl, FIPACONSTANTS.INFORM, getKeywords() );
               done();
            }
         }
         else if (function.equals(IHCONSTANTS.KEYWORD_LIST))
         {
            if (n == 1) 
            {
               sendReply(_acl, FIPACONSTANTS.AGREE); 
               sendReply(_acl, FIPACONSTANTS.INFORM, getKeywords() );
               done();
            }
         }
         else if (function.equals(IHCONSTANTS.REGISTERUI))
         {
            // set the UIAgent
            if (n == 2) 
            {
               String onoff = st.nextToken(); 
               if ( onoff.equals(IHCONSTANTS.ONLINE) ) 
               { 
                  setUIAgent(_acl.getSenderAID());
                  setStatus(true);
               }
               else if ( onoff.equals(IHCONSTANTS.OFFLINE) ) 
               {
                  setUIAgent(null);
                  setStatus(false);
               }
               sendReply(_acl, FIPACONSTANTS.AGREE); 
               sendReply(_acl, FIPACONSTANTS.INFORM, getStrStatus() );
               done();
            }
         }
         else if (function.equals(IHCONSTANTS.HELP))
         {
            // someone needs help
            if (n == 3) 
            {
               sendReply(_acl, FIPACONSTANTS.AGREE); 
               String person = st.nextToken();
               String kw = st.nextToken();
               
               if ( _status && (hasKeyword(kw)))
                  sendReply(_acl, FIPACONSTANTS.INFORM,  IHCONSTANTS.YES);
               else 
                  sendReply(_acl, FIPACONSTANTS.INFORM,  IHCONSTANTS.NO);
               done();
            }
         }
         else if (function.equals(IHCONSTANTS.FIND))
         {
            // find a helper
            if (n == 3) 
            {
               sendReply(_acl, FIPACONSTANTS.AGREE); 
               String person = st.nextToken();
               String kw = st.nextToken();
               newTask( new FindHelpTask( kw, person ), IHCONSTANTS.FIND_TIMEOUT); 
            }
         } 
         else 
         {
            sendNotUnderstood(_acl); 
            done();
         } 
      }


      public void donePersAgent_FindHelpTask(Object o) 
      {
         String helpers = (String) o;
         sendReply(_acl, FIPACONSTANTS.INFORM, helpers);
         done();
      } 


      public void timeoutPersAgent_FindHelpTask(Task t) 
      {
         // send message back saying could not find any helpers! 
         sendReply(_acl, FIPACONSTANTS.INFORM, "No HELPers Found");
         done();
      } 

   } // HandleRequestTask


   /** Find a helper. Send a message to the MM and wait for a reply
   */
   public class FindHelpTask extends Task 
   {

      private String _kw; 
      private String _person; 


      public FindHelpTask(String kw, String person) 
      {
         _kw = kw; 
         _person = person;
      }


      protected void startTask()
      {
         ACL req = new ACL();
         req.setSenderAID( getAID());
         // send message to proxy and add real addressee to replyto
         req.setReceiverAID( _matchmaker); 
         //
         req.setPerformative( FIPACONSTANTS.REQUEST );
         req.setOntology( IHCONSTANTS.IHELP_ONT );
         req.setContentObject( IHCONSTANTS.FIND + " " +_person + " " + _kw);
         req.setProtocol(FIPACONSTANTS.FIPA_REQUEST);
         forward( req );
      }

      public void handleAgree(Conversation c)
      {
      }


      public void handleInform(Conversation c)
      {
         ACL reply = c.getACL( c.getLatestMessageIndex());
         String helpers = (String) reply.getContentObject();
         done(helpers);
      }

   } // FindHelpTask


   public static void main( String args[] ) {

      if ( args == null || args.length < 2 ) {

         System.out.println( "Usage: java fipaos.tutorial.PersAgent <platform.profile> <name>" );
         return;
      }

      new PersAgent( args[0], args[1], "FIPA-OS" );

   } // main()

} // PersAgent
