/* 
   Agent Server AgentServer V1.0, November 6, 2001  

   Written by: 
      Alan Fedoruk                                           -- amf673@mail.usask.ca
      MADMUC Lab                                             -- http://bistrica.usask.ca/madmuc/index.htm
      Dept. of Computer Science, University of Saskatchewan. -- http://www.cs.usask.ca

   Copyright (c) 2001 Alan M. Fedoruk
*/

package agentserver;

// Import the fipa-os classes
import fipaos.agent.*;
import fipaos.agent.conversation.*;
import fipaos.agent.task.*;
import fipaos.mts.*;
import fipaos.ont.fipa.*;
import fipaos.ont.fipa.fipaman.*;
import fipaos.parser.acl.ACLMessage;
import fipaos.util.DynamicClass;

// Import registration exception classes
import fipaos.platform.ams.AMSRegistrationException;
import fipaos.platform.df.DFRegistrationException;
import fipaos.util.DIAGNOSTICS;

import java.util.*;

import localpostoffice.*;


/** AgentServer implements an AgentServer for FIPAOS. 

    <P>
    An AgentServer provides a lightweight process space, which can 
    be used to host other agents.  The AgentServer extends the FIPAOSAgent shell. 
    <P>
    All requests are sent using the agentserver-management ontology. Any other 
    ontology is ignored.  The possible requests are: 

    <UL>
       <LI>create      - create an agent. 
       <LI>list        - list the agents that I have already created. 
       <LI>properties  - list my properties.
    </UL>
    When a create request is recieved, or a create, agents are created in this process space. 
    <P>
    The list request will return a list of the agents currently running in this AgentServer.
    <P>
    The properties request will return the name of this AgentServer and the number of 
    agents currently hosted by this AgentServer, and the maximum number of agents 
    this agentserver can host.  
    <P>
    An agentserver can be run with or without a gui, depending on command line switches. 
    <P> 
    The agentserver has a limit of the number of agents that it can start. This is settable 
    via command line swtiches. 
    <P>
    Usage: 
    <BR>
       java fipaos.agentserver.AgentServer <platform.profile> <name> <max_agents> <gui|nogui>
        
*/
public class AgentServer extends FIPAOSAgent implements LocalMsg
{

   /** This Agent type */
   private static final String AGENT_TYPE = ASCONSTANTS.AGENTSERVER;

   /** List of agents that this agentserver has started */
   private Hashtable _started_agents = new Hashtable(); 

   /**  GUI for this AgentServer */
   private AgentServerGUI _gui;

   /** Location of platform profile for use when creating agents */
   private String _platform_profile;

   /** The maximum number of agents that this AgentServer can start */
   private int _max_agents = ASCONSTANTS.MAX_AGENTS;

   private boolean _started; 

   /** if desired, can use the localpostoffice feature to speed up communication between local agents. */     
   private LocalPostOffice _localpo = null; 

   /** constructor - sets up an IdleTask as the default
       ListenerTask for incoming messages, and starts off registration
       with the AMS and DF
   */
   public AgentServer (String platform, String name, String ownership, String ma, String gui ) 
   {
   
      super( platform, name, ownership );
    
      _platform_profile = platform;
      _max_agents = Integer.parseInt(ma);      

      /* create the localpostoffice */
      _localpo = new LocalPostOffice( _started_agents);

      // Create a GUI for this Agent so we can see what is going on.
      if ( gui.equals("gui"))
      {
         try 
         {
            _gui = new AgentServerGUI( getAID().getName(), this );
            // _gui.refresh();
         } 
         catch ( Throwable t ) 
         { 
            System.out.println("Cannot Create GUI!");
         }
      }
    
      // Instantiate a new IdleTask to receieve all incoming 
      // communications which are not associated with any other Task

      setListenerTask( new IdleTask() );
  
      // Register with the local AMS (i.e. the platform)
      try 
      {

         registerWithAMS();
         writeMsg("Registered with AMS" );

      } 
      catch ( AMSRegistrationException amsre ) 
      {
         String reason = amsre.getExceptionReason();
      
         if ( reason == null || !reason.equals( FIPAMANCONSTANTS.AGENT_ALREADY_REGISTERED)) 
         {
            shutdown();
            return;
         }
         writeMsg("Already Registered with AMS" );
      
      } // end catch 
    
      // Register with the local DF
      try 
      {
         registerWithDF( AGENT_TYPE );
         writeMsg("Registered with DF" );
      } 
      catch (DFRegistrationException dfre) 
      {
         String reason = dfre.getExceptionReason();
      
         if ( reason == null || !reason.equals( FIPAMANCONSTANTS.AGENT_ALREADY_REGISTERED) ) 
         {
            shutdown();
            return;
         }
         writeMsg("Already Registered with DF" );
      } // end catch

   } // AgentServer()
  

   /** Invoked when this Agent is about to be shutdown - this overrides the
       shutdown() method in the FIPAOSAgent class to enable the AgentServerGUI
       associated with this AgentServer to be removed
   */
   public synchronized void shutdown() 
   {
      if (registeredWithDF() == true) 
      {
         // Attempt to deregister with DF
         try 
         {
            deregisterWithDF();
         } 
         catch (DFRegistrationException dfre) 
         {
            // A problem deregistering occured..... 
            String reason = dfre.getExceptionReason();

         } // end catch
      } // if 
    
      if (registeredWithAMS() == true) 
      {
         // Attempt to deregister with AMS
         try 
         {
            deregisterWithAMS();

         } 
         catch (AMSRegistrationException amsre) 
         {
            // A problem deregistering occured.....
            String reason = amsre.getExceptionReason();
     
         } // catch
      } // if 
     
      // Dispose of the GUI
      if( _gui != null ) 
      {
         _gui.dispose();
      }

      // send a shutdown to any agents started in this process. 

      for (Enumeration e = _started_agents.keys(); e.hasMoreElements() ;) 
      {
         FIPAOSAgent ra = (FIPAOSAgent)_started_agents.get( (String)e.nextElement());
         ra.shutdown();
      }

      // Now call shutdown in the agent shell to release the core components
      super.shutdown();

   } // shutdown()
  

   /** This method is invoked when GUI should be shown - generally
       this is invoked by the AgentLoader when the Agent is in the List
       of running Agents - the default implementation in FIPAOSAgent does 
       nothing.
   */
   public void activate() 
   {
      // Check if we've got a GUI, if so, show it
      if( _gui != null ) 
      {
         _gui.setVisible( true );
      }

   } // activate()
    

   /** set the localpostoffice to the given argument */
   public void setLocalPostOffice( LocalPostOffice localpo)
   {
      _localpo = localpo;
   }

   /** return the current localpostoffice object */
   public LocalPostOffice getLocalPostOffice( )
   {
      return _localpo;
   }

   /** write a message, either to the GUI or to system.out
   */
   public void writeMsg( String msg)
   {
      if( _gui != null ) 
         _gui.write( msg);
      else 
         System.out.println( msg);
   }


   /** Send a reply message in response to the given ACL message
       @param acl The ACL object to reply to
       @param performative The performative 
       @param content Any additional message
   */
   public boolean sendReply( ACL orig_acl, String performative, String content )
   {
      boolean done = false;
      try
      {
         // Duplicate ACL
         ACL acl = (ACL) orig_acl.clone();
            
         // Setup message
         acl.setPerformative( performative );
         acl.setReceiverAID( acl.getSenderAID() );
         acl.setSenderAID( getAID() );
         acl.setInReplyTo( acl.getReplyWith() );
         acl.setReplyWith( null );
            
         // Set content
         acl.setContentObject( content );
            
         // Attempt to send the agree
         forward( acl );
            
         // We're done
         done = true;
      }
      catch( Throwable t )
      {
         DIAGNOSTICS.println( "Error occured when sending agree", t, this,
         DIAGNOSTICS.LEVEL_MAX );
      }
      return done;

   } // sendReply()
	

   public boolean sendReply( ACL orig_acl, String performative )
   {
      return sendReply(orig_acl, performative, "");
   }

   /** Starts an agent
      @param agent_name Name of the Agent to start
      @param class_name Name of the class for this Agent
      @param owner Owner of the Agent
   */
   public synchronized boolean startAgent( final String class_name, final String agent_name, 
         final String owner, final String[] parameters) 
   {
      _started = true;

      // Check that no agents with the same name are currently alive -- need to check whole platform!!!!
      if ( _started_agents.containsKey( agent_name ) ) 
      {
         writeMsg( "Agent \"" + agent_name + "\" is already running!" );
         return false;
      }

      // Check if name is valid (contains no spaces!)
      if( agent_name.indexOf( " " ) != -1 ) 
      {
         writeMsg( "Agent names cannot contain spaces!" );
         return false;
      }

      // A thread group so that the threads belonging to an agent can be kept track of
      ThreadGroup new_tg = new ThreadGroup( agent_name );

      // Start agent in new thread
      Thread t = new Thread( new_tg, new Runnable() 
      {
         public void run() 
         {
            try 
            {
               // Add place-holder to list of active agents
               // _started_agents_list.add( agent_name );
               _started_agents.put( agent_name, new Object() );

               // Setup expected parameters
               Class args[] = new Class[parameters.length + 3]; 
               for ( int i=0; i < args.length; i++) 
               { 
                  args[i] = String.class;
               }

               Object params[] = new Object[args.length];
               params[0] = _platform_profile;
               params[1] = agent_name; 
               params[2] = owner; 
               for ( int i=3; i < params.length; i++) 
               {
                  params[i] = parameters[i-3]; 
               } 

               DynamicClass dc = new DynamicClass( class_name, args, params );

               // Add real FIPAOSAgent instance to map of started Agents
               _started_agents.put( agent_name, dc.getObject() );
            } 
            catch (Throwable x) 
            {
               writeMsg("Could Not Start agent " + agent_name);
               x.printStackTrace();
               _started_agents.remove( agent_name );
               _started = false;
            } 
         } // run()
      } 
      );

     t.start();
     return _started;

   } //StartAgent()


  
   /** Create an agent
   */
   public class CreateTask extends Task
   {
      private ACL          _acl; 
      private String       _agent_name;
      private String       _class_name;


      public CreateTask( ACL acl) 
      {
         _acl = acl; 
      } // CreateTask()


      protected void startTask ()
      {
         if ( _started_agents.size() < _max_agents) 
         {
            sendReply(_acl, FIPACONSTANTS.AGREE, "");

            String content = (String)_acl.getContentObject();  
            StringTokenizer st = new StringTokenizer(content);
            st.nextToken();            
            String class_name = st.nextToken();
            String agent_name = st.nextToken();
            String owner = st.nextToken();
            String params[] = new String[st.countTokens()];

            for ( int i=0; i < st.countTokens(); i++) 
            {
               params[i] = st.nextToken();
            }

            if ( startAgent(class_name, agent_name, owner, params)) 
            {
               // agent was started. 
               sendReply(_acl, FIPACONSTANTS.INFORM, "Agent Started OK");
               writeMsg("Started Agent " + agent_name + " of type " + class_name);
            } 
            else 
            {
               sendReply(_acl, FIPACONSTANTS.FAILURE, "Could not start Agent");
               writeMsg("Could not start Agent " + agent_name + " of type " + class_name);
            }
            
         }
         else 
         {
            sendReply(_acl, FIPACONSTANTS.REFUSE, "No More Agents");
            writeMsg("No more free agent slots.");
         }
         done();
      } // startTask()

   } // CreateTask


   /** IdleTask. Handles all processing of messages not associated with any 
       other conversation. 
   */
   public class IdleTask extends Task 
   {

      public IdleTask() 
      {
      } // IdleTask()
    
    
      protected void startTask() 
      {
      } // StartTask()


      public void handleRequest( Conversation conv) 
      {
         // Get the type of request by checking ontology 
         // if it is not repserver-management then respond with not understood. 

         ACL acl = conv.getACL( conv.getLatestMessageIndex() );

         if( acl.getOntology() == null || !acl.getOntology().equalsIgnoreCase( ASCONSTANTS.AGENTSERVER_ONT ) ) 
         {
	    sendNotUnderstood( acl );
	    done();
         }

         try 
         {  // Get content of message
	    String content = (String)acl.getContentObject();
            StringTokenizer st = new StringTokenizer(content); 

            String function = "";
	
            if ( st.hasMoreTokens() ) 
            {
               function = st.nextToken(); 
            } 
            else 
            {
               sendNotUnderstood(acl);  
               return;
            }

            // Decide which sub-tasks to spawn

            if ( function.equals( ASCONSTANTS.CREATE) ) 
            {
	       newTask( new CreateTask( acl), conv );
            } 
            else if ( function.equals( ASCONSTANTS.LIST ) ) 
            {
	       newTask( new ListTask( acl ), conv );
            } 
            else if ( function.equals( ASCONSTANTS.PROPERTIES ) ) 
            {
	       newTask( new PropertiesTask( acl ), conv ); 
            } 
            else 
            {
               sendNotUnderstood( acl );
            }
         } 
         catch( Throwable t ) 
         {
            sendNotUnderstood( acl );
         } 
      } // handleRequest()
      

      public void doneAgentServer_CreateTask( Task t) 
      {
      }

    
      public void doneAgentServer_ListTask(Task t) 
      {
      }


      public void doneAgentServer_PropertiesTask(Task t) 
      {
      }

   } // IdleTask()


   /** List the agents that have been created in this AgentServer. 
   */
   public class ListTask extends Task
   {
      private ACL          _acl; 
      private String       _agent_name;
      private String       _class_name;
      private int          _N;


      public ListTask( ACL acl) 
      {
         _acl = acl; 
      } // 


      protected void startTask ()
      {
         sendReply(_acl, FIPACONSTANTS.AGREE, "" );  
         String agent_list = "";
         for (Enumeration e = _started_agents.keys(); e.hasMoreElements();) 
         {
            agent_list = agent_list + " " + e.nextElement(); 
         } // for
  
         sendReply( _acl, FIPACONSTANTS.INFORM, agent_list);
         done();

      } // startTask()

   } // ListTask
 

   /** List the properties of this AgentServer. 
   */
   public class PropertiesTask extends Task
   {
      private ACL          _acl; 
      private String       _agent_name;
      private String       _class_name;


      public PropertiesTask( ACL acl) 
      {
         _acl = acl; 
      } // 


      protected void startTask ()
      {
         String response;
         int rem_agents = _max_agents - _started_agents.size();
         sendReply(_acl, FIPACONSTANTS.AGREE, "");
         response = "Name: " + getAID().getName(); 
         response = response + " Max Agents: " + _max_agents;
         response = response + " Remaining Agents: " + rem_agents;

         sendReply(_acl, FIPACONSTANTS.INFORM, response);
         done();
      }

   } // PropertiesTask


   /** Test-harness
   */
   public static void main( String args[] ) 
   {
      if ( args == null || args.length < 3 ) 
      {
         System.out.println( 
               "Usage: java fipaos.agentserver.AgentServer <platform.profile> <name> <max_agents> <gui|nogui>" );
         return;
      }
      new AgentServer( args[0], args[1], "FIPA-OS", args[2], args[3] );

   } // main()

} // AgentServer
