/* 
   IHELP Simulation - MatchMaker Agent V1.0 

   Written by: 
      Alan Fedoruk                                           -- amf673@mail.usask.ca
      MADMUC Lab                                             -- http://bistrica.usask.ca/madmuc/index.htm
      Dept. of Computer Science, University of Saskatchewan. -- http://www.cs.usask.ca

   Copyright (c) 2001 Alan M. Fedoruk
*/

package ihelp;

import fipaos.agent.*;
import fipaos.agent.conversation.*;
import fipaos.agent.task.*;
import fipaos.mts.*;
import fipaos.ont.fipa.*;
import fipaos.ont.fipa.fipaman.*;
import fipaos.parser.*;
import fipaos.parser.acl.ACLMessage;
import fipaos.util.DIAGNOSTICS;
import fipaos.platform.*;
import fipaos.platform.ams.*;
import fipaos.platform.df.*;

import repserver.*;

import java.util.*;
import java.io.*;

/** MMAgent is the IHelp matchmaker agent

    It maintains a hashtable of personal agents, broadcasts help requests 
    to all the PersAgents that it knows about and assembles the replies 
    that it gets.  */
public class MMAgent extends RepAgent 
{

   /* my type */
   private static final String AGENT_TYPE = IHCONSTANTS.MMAGENT;

   /* Hashtable of the agents that I know about. */
   private Hashtable _agents; 

   /* A random number generator for simulated failures */
   private Random _r = new Random();


   public MMAgent( String platform, String name, String ownership, AgentID rs, String role, String group_name ) 
   {
   
      super( platform, name, ownership, rs, role, group_name );

      _agents = new Hashtable();

      setListenerTask( new IdleTask());

   } // MMAgent()
  

   public void goActive() 
   {
      // no action needed when MMAgent goes from REPLICANT to ACTIVE_AGENT
   }


   public synchronized void shutdown() 
   {
      super.shutdown();
   } // shutdown()

    

   /** true if the given AgentID is registered with this MM
   */
   public boolean hasPersAgent( AgentID aid) 
   {
      if (_agents.containsKey( aid.getName()))
         return true; 
      else 
         return false; 
   }


   /** Add the given AgentID to the MM
   */
   public void addPersAgent( AgentID aid) 
   {
      delPersAgent(aid); 
      _agents.put( aid.getName(), aid);
   }


   /** Remove the given AgentID from the MM
   */
   public void delPersAgent( AgentID aid) 
   {
      if (_agents.containsKey( aid.getName()))
         _agents.remove(aid.getName());
   }


   /** List, as a string the names of all the agents 
       currently registered with this matchmaker
   */
   public String listAgents() 
   {
      String res = "";
      for (Enumeration e = _agents.keys(); e.hasMoreElements(); )
      {
         res = res + " " + e.nextElement();
      }
      return res;
   }


   /** generate a random number, and if the number is in the 
       range to fail, shutdown this agent. 
   */
   public void checkForFailure()
   {
      float r = _r.nextFloat();
      if ( r <= IHCONSTANTS.FAIL_RATE) 
      {
         System.out.println(getAID().getName() + " is failing ");
         shutdown();
      }
   }


   public byte[] getAState() 
   {
      try 
      {
         ByteArrayOutputStream bos = new ByteArrayOutputStream();
         ObjectOutputStream oos = new ObjectOutputStream(bos);
         // DIAGNOSTICS.println("Get Agent State ", DIAGNOSTICS.LEVEL_4);

         // write all the variables that define this agent 
         oos.writeObject(_agents);
   
         bos.close();
         oos.close();
         return ( bos.toByteArray());
      } 
      catch ( Exception e) 
      {
         DIAGNOSTICS.println("Failed to Get Agent State", DIAGNOSTICS.LEVEL_4);
         return ( new byte[1]);
      }

   } 


   /** set the state of the agent, given a byte stream.  */
   public boolean setAState(byte[] b) 
   {
      try 
      {
         ByteArrayInputStream bis = new ByteArrayInputStream(b);
         ObjectInputStream ois = new ObjectInputStream(bis);

         _agents = (Hashtable) ois.readObject();

         DIAGNOSTICS.println("Set Agent State " , DIAGNOSTICS.LEVEL_4);

         ois.close();
         bis.close();
      } 
      catch ( Exception e) 
      {
         DIAGNOSTICS.println("Failed to Set Agent State", DIAGNOSTICS.LEVEL_4);
         e.printStackTrace();
         return false;
      }
      return true;

   } // setAState


   /** IdleTask 
   */
   public class IdleTask extends Task 
   {

      public IdleTask() 
      {
      } // IdleTask()
    
    
      protected void startTask() 
      {
      } // StartTask()


      public void handleRequest( Conversation conv ) 
      {
         // System.out.println("MMAgent: " + getAID().getName() + " Tasks: " + _tm.getTaskCount() + " Convs: " + getConvs() );
         // Get the type of request
         ACL acl = conv.getACL( conv.getLatestMessageIndex() );

         // Check ontology if this is repserver ont, pass request to RepServerManagementTask
         if ( acl.getOntology().equalsIgnoreCase( RSCONSTANTS.REPSERVER_ONT ) ) 
         {
            newTask ( new RepServerManagementTask(acl), conv); 
	 } // if
         else if ( isActiveAgent() ) 
         {
            // if this is the active agent, process the request.
            if ( acl.getOntology().equalsIgnoreCase( IHCONSTANTS.IHELP_ONT ) ) 
            {
               newTask ( new HandleRequestTask(conv, acl), conv); 
            } // if
         } // else if
      } // handleRequest()


      public void doneRepAgent_RepServerManagementTask(Task t) 
      {
      }


      public void doneMMAgent_HandleRequestTask(Object o)
      {
         String function = (String) o;
         // after every request
         checkForFailure();
         // if the function was something that changed my state, update my state
         if ( function.equals(IHCONSTANTS.REGISTER) || function.equals(IHCONSTANTS.DEREGISTER))
            newTask( new StateTask());
      }
    

      public void doneRepAgent_StateTask(Task t)
      {
      }
    

      public void timeoutRepAgent_StateTask(Task t)
      {
      }

   } // IdleTask()


   /** Handles all incoming i-help requests
   */
   public class HandleRequestTask extends Task
   {


      private ACL _acl;
      private Conversation _conv;
      // The operation performed
      private String _function; 
      

      public HandleRequestTask ( Conversation conv, ACL acl) 
      {
         _acl = acl;
         _conv = conv;
      }
      

      public void startTask() 
      {

         String content = (String)_acl.getContentObject();
         StringTokenizer st = new StringTokenizer(content);
         int n = st.countTokens();
         _function = st.nextToken(); 

         // register the sender personal agent with this MM. 
         if (_function.equals(IHCONSTANTS.REGISTER))
         {
            String aidString = 
                  content.substring( content.indexOf(IHCONSTANTS.REGISTER) + IHCONSTANTS.REGISTER.length());
            sendReply(_acl, FIPACONSTANTS.AGREE);
            try 
            {
               AgentID aid = new AgentID( aidString);
               addPersAgent(aid); 
               sendReply(_acl, FIPACONSTANTS.INFORM);
               done(_function);
            }
            catch (ParserException p)
            {
               sendReply(_acl, FIPACONSTANTS.FAILURE, "Malformed AgentID");
               done(_function);
            }
         }
         // deregister the sender personal agent with this MM. 
         else if (_function.equals(IHCONSTANTS.DEREGISTER))
         {
            if (n == 1) 
            {
               sendReply(_acl, FIPACONSTANTS.AGREE);
               AgentID aid = _acl.getSenderAID();
               if ( hasPersAgent(aid)) 
               {
                  delPersAgent(aid); 
               }
               sendReply(_acl, FIPACONSTANTS.INFORM);
            }
         }
         // find helpers capable of the keyword given. 
         else if (_function.equals(IHCONSTANTS.FIND))
         {
            // find helpers capable of the keyword given. 
            // later can add capability for more keywords.
            if (n == 3) 
            {
               sendReply(_acl, FIPACONSTANTS.AGREE);
               String person = st.nextToken();
               String kw = st.nextToken();
               newTask ( new FindTask(kw, person), IHCONSTANTS.FIND_TIMEOUT); 
            }
         }
         else 
         {
            sendNotUnderstood(_acl); 
            done(_function);
         } 
      }


      public void doneMMAgent_FindTask( Object o) 
      {
         sendReply(_acl, FIPACONSTANTS.INFORM, (String) o);
         done(_function);
      // send inform with results of search. 
      }


      public void timeoutMMAgent_FindTask( Task t) 
      {
         sendReply(_acl, FIPACONSTANTS.INFORM, "");
         done(_function);
      }


   } // HandleRequestTask


   /** FindTask sends a help request to all the personal agents that 
       this MM knows about. The responses are tabulated and sent 
       back to the personal agent that asked for help. 
   */
   public class FindTask extends Task 
   {

      /** count of how many agents I have to send to. 
      */
      private int _count; 
      private String _response = "";
      private String _keyword; 
      private String _person; 


      public FindTask (String kw, String person)
      {
         _keyword = kw;
         _person = person;
      } // FindTask()


      protected void startTask() 
      {
         _count = _agents.size();
         for (Enumeration e = _agents.keys(); e.hasMoreElements() ;) 
         {
            AgentID aid = (AgentID)_agents.get((String)e.nextElement());
            if ( !aid.getName().equals(_person) ) 
            { // dont ask myself!
               newTask( new SendFindTask( aid, _keyword, _person), IHCONSTANTS.FIND_TIMEOUT); 
            }
            else 
               _count--;
         }
      } // startTask()


      public void doneMMAgent_SendFindTask( Task t)
      {
         if ( t instanceof SendFindTask )
         {
            SendFindTask s = (SendFindTask) t;
            String response = s.getResponse(); 
            String aname = s.getAgent().getName();
            if ( response.equals("YES")) 
            { 
               _response = _response + " " + aname;
            }
         }
         decrementCount();
      }


      public void timeoutMMAgent_SendFindTask( Task t)
      {
         decrementCount();
      }


      public void errorMMAgent_SendFindTask( Task t, Throwable th)
      {
         decrementCount();
      } 


      private synchronized void decrementCount()
      {
         _count--; 
         if ( _count <= 0)
         {
            done(_response);
         }

      } // decrementCount()


   } // FindTask



   /** send a message to the given agent to see if the 
       agents user is willing and able to help. 
   */
   public class SendFindTask extends Task
   {


      /* the agent to send the request to */
      private AgentID _aid; 

      /* the keyword to ask for */
      private String _keyword; 

      /* the person asking for help */
      private String _person; 
      
      /* the response from this agent. Set to NO by default */
      private String _response = IHCONSTANTS.NO; 


      /** return the AgentID of the agent we are asking 
      */
      public AgentID getAgent() 
      {
         return _aid; 
      }


      /** return the response that we got. 
      */
      public String getResponse()
      {
         return _response;
      } 


      public SendFindTask(AgentID aid, String kw, String person) 
      {
         _aid = aid; 
         _keyword = kw;
         _person = person;
      }


      protected void startTask()
      {
         // send the request. Since we are initiated this conversation 
         // and this is a replicated agent... we need to do a little extra  
         ACL acl = new ACL();
         acl.setPerformative( FIPACONSTANTS.REQUEST );
         acl.setSenderAID( _owner.getAID() );
         // send the message to the group_proxy
         acl.setReceiverAID( getProxyAID() );
         // add the agent to forward to in the replyto
         acl.addReplyToAID( _aid);
         acl.setContentObject( IHCONSTANTS.HELP + " " + _person + " " + _keyword );
         acl.setOntology( IHCONSTANTS.IHELP_ONT );
         acl.setProtocol( FIPACONSTANTS.FIPA_REQUEST );
         forward( acl );
      }


      public void handleAgree(Conversation conv) 
      {
         // ignore the agree
      } 


      public void handleInform(Conversation conv) 
      {
         ACL acl = conv.getACL( conv.getLatestMessageIndex() );
         _response = (String) acl.getContentObject(); 
         done();
      } 

   } // SendFindTask


   /** Test harness */
   public static void main( String args[] ) {

      if ( args == null || args.length < 2 ) {

         System.out.println( "Usage: java fipaos.tutorial.MMAgent <platform.profile> <name>" );
         return;
      }

      new MMAgent( args[0], args[1], "FIPA-OS", null, "REPLICANT", "NONE" );

   } // main()

} // MMAgent
