/* 
   IHELP Simulation - PersAgent V1.0 

   Written by: 
      Alan Fedoruk                                           -- amf673@mail.usask.ca
      MADMUC Lab                                             -- http://bistrica.usask.ca/madmuc/index.htm
      Dept. of Computer Science, University of Saskatchewan. -- http://www.cs.usask.ca

   Copyright (c) 2001 Alan M. Fedoruk
*/

package ihelp;

// Import the fipa-os classes
import fipaos.agent.*;
import fipaos.agent.conversation.*;
import fipaos.agent.task.*;
import fipaos.mts.*;
import fipaos.ont.fipa.*;
import fipaos.ont.fipa.fipaman.*;
import fipaos.parser.acl.ACLMessage;
import fipaos.platform.ams.AMSRegistrationException;
import fipaos.platform.df.DFRegistrationException;
import fipaos.util.DIAGNOSTICS;

// Import java classes
import java.util.*;
import java.io.*;

// Import the replication server classes
import repserver.*;

/** A users Personal Agent. 
    Stores a little about a user and knows if the user is on line 
    or off.  */
public class PersAgent extends RepAgent 
{

   /* This Agent type */
   private static final String AGENT_TYPE = IHCONSTANTS.PERSAGENT;

   /* The AID of the UI agent associated with this Personal Agent */
   private AgentID _uiagent = null;

   /* The name of my user */
   private String _user_name;  

   /* A vector of the things my user knows about */
   private Vector _keywords; 

   /* The status of my user true = online, false = offline */
   private boolean _status; 

   /* A matchmakers agent */
   private AgentID _matchmaker;

   /* A random number generator for simulated failures */
   private Random _r = new Random();
  

   /** standard RepAgent constructor */
   public PersAgent( String platform, String name, String ownership, 
         AgentID rs, String role, String group_name ) 
   {
   
      super( platform, name, ownership, rs, role, group_name );
    
      // Instantiate a new IdleTask to receieve all incoming 
      // communications which are not associated with any-other Task

      setListenerTask( new IdleTask() );
  
      setUIAgent(null); 
      setUserName(null);  
      _keywords = new Vector(); 
      // add the java keyword as a default keyword
      _keywords.addElement("java");
      setStatus(false); 

   } // PersAgent()
  

   /** when going to ACTIVE_AGENT, look for a new MMAgent */
   public void goActive() 
   {
      _tm.newTask( new FindMMTask() ); 
   }


   public void donePersAgent_FindMMTask(Task t)
   {
   }


   /** Invoked when this Agent is about to be shutdown - this overrides the
       shutdown() method in the FIPAOSAgent class to enable the CountAgentGUI
       associated with this CountAgent to be removed */
   public synchronized void shutdown() 
   {
      // Now call shutdown in the agent shell to release the core components
      super.shutdown();

   } // shutdown()


   /** set the user_name for the agent */
   public void setUserName(String name) 
   {
      _user_name = name; 
   } // setUserName()


   /** get the user_name for the agent */
   public String getUserName() 
   {
      return _user_name; 
   } // getUserName()


   /** set the AgentID of the user interface agent */
   public void setUIAgent(AgentID aid) 
   {
      _uiagent = aid; 
   } // setUIAgent()


   /** get the AgentID of the user interface agent */
   public AgentID getUIAgent() 
   {
      return (_uiagent); 
   } // getUIAgent()


   /** add the given keyword to the list of keywords */
   public void addKeyword(String kw) 
   {
      if (!hasKeyword(kw)) 
         _keywords.addElement(kw);
   } // addKeyword()


   /** get a string of the list of keywords */
   public String getKeywords() 
   {
      String res = "";
      for ( int i=0; i < _keywords.size(); i++) 
      {
         res = res + " " + (String)_keywords.elementAt(i);
      }
      return res;
   }


   /** return true if the keyword is in the list. */ 
   public boolean hasKeyword( String kw) 
   {
      return _keywords.contains(kw); 
   } // hasKeyword()


   /** set the status, true is user is available, false otherwise. */ 
   public void setStatus(boolean st) 
   {
      _status = st;
   } // setStatus()


   /** get the status. */
   public boolean getStatus() 
   {
      return(_status);
   } // setStatus()

   /** get the String status. */
   public String getStrStatus() 
   {
      if ( getStatus()) 
         return ( IHCONSTANTS.ONLINE);
      else 
         return ( IHCONSTANTS.OFFLINE);
   } // setStatus()


   /** return the state of this agent.  */
   public byte[] getAState() 
   {
      try 
      {
         ByteArrayOutputStream bos = new ByteArrayOutputStream();
         ObjectOutputStream oos = new ObjectOutputStream(bos);
         // DIAGNOSTICS.println("Get Agent State ", DIAGNOSTICS.LEVEL_4);

         // write all the variables that define this agent 
         oos.writeObject(_uiagent);
         oos.writeObject(_user_name);
         oos.writeObject(_keywords);
         oos.writeBoolean(_status);
         oos.writeObject(_matchmaker);
   
         bos.close();
         oos.close();
         return ( bos.toByteArray());
      } 
      catch ( Exception e) 
      {
         DIAGNOSTICS.println("Failed to Get Agent State", DIAGNOSTICS.LEVEL_4);
         return ( new byte[1]);
      }

   } 


   /** set the state of the agent, given a byte stream.  */
   public boolean setAState(byte[] b) 
   {
      try 
      {
         ByteArrayInputStream bis = new ByteArrayInputStream(b);
         ObjectInputStream ois = new ObjectInputStream(bis);

         _uiagent = (AgentID) ois.readObject();
         _user_name = (String) ois.readObject();
         _keywords = (Vector) ois.readObject();
         _status = ois.readBoolean();
         _matchmaker = (AgentID) ois.readObject();

         DIAGNOSTICS.println("Set Agent State " , DIAGNOSTICS.LEVEL_4);

         ois.close();
         bis.close();
      } 
      catch ( Exception e) 
      {
         DIAGNOSTICS.println("Failed to Set Agent State", DIAGNOSTICS.LEVEL_4);
         e.printStackTrace();
         return false;
      }
      return true;

   } // setAState


   /** generate a random number, and if the number is in the 
       range to fail, shutdown this agent.  */
   public void checkForFailure()
   {
      float r = _r.nextFloat();
      if ( r <= IHCONSTANTS.FAIL_RATE) 
      {
         System.out.println(getAID().getName() + " is failing ");
         shutdown();
      }
   }



   /** IdleTask 
   */
   public class IdleTask extends Task 
   {


      public IdleTask() 
      {
      } // IdleTask()
    
    
      protected void startTask() 
      {
      } // StartTask()


      public void handleRequest( Conversation conv ) 
      {
         // System.out.println( "PersAgent " + getAID().getName() + " Tasks: " + _tm.getTaskCount() + " Convs: " + getConvs());
         // Get the type of request
         ACL acl = conv.getACL( conv.getLatestMessageIndex() );

         // Check ontology
         if ( acl.getOntology().equalsIgnoreCase( RSCONSTANTS.REPSERVER_ONT ) ) 
         {
            newTask ( new RepServerManagementTask(acl), conv); 
	 } 
         else if ( isActiveAgent() ) 
         {
            // if this is the active agent, process the request.
            if ( acl.getOntology().equalsIgnoreCase( IHCONSTANTS.IHELP_ONT ) ) 
            {
               newTask ( new HandleRequestTask(acl, conv), conv); 
            }
         }
      } // handleRequest()


      public void doneRepAgent_RepServerManagementTask(Task t) 
      {
      }


      public void donePersAgent_HandleRequestTask(Object o)
      {
         String function = (String)o;
         checkForFailure();
         if ( function.equals(IHCONSTANTS.KEYWORD_ADD) || function.equals(IHCONSTANTS.REGISTERUI) )  
            // if the operation was one that changed my state
            // the update my state with the replication server
            newTask( new StateTask());
      }
    
      public void doneRepAgent_StateTask(Task t)
      {
      }
    
      public void timeoutRepAgent_StateTask(Task t)
      {
      }

   } // IdleTask()



   public class FindMMTask extends Task 
   {

      public FindMMTask()
      {
      }

 
      protected void startTask() 
      {
	 // Search the DF for a MM
         DFAgentDescription df_desc = new DFAgentDescription();
         // Create a ServiceDescription of the match-maker agents
         ServiceDescription sd = new ServiceDescription();
         sd.setServiceName(IHCONSTANTS.MMAGENT);
         sd.setServiceType(IHCONSTANTS.MMAGENT); 		
         df_desc.setAgentServices( new ServiceDescription[] { sd } );
			
         // Start a search for other Agents
         searchDF( df_desc );
         DIAGNOSTICS.println( "PersAgent: Started DF search", DIAGNOSTICS.LEVEL_4 );
			

         // register with an MM 
         
      } // startTask()


      /** This method is invoked when a DFSearchTask has completed
          succesfully.
      */ 
      public void DFSearchResults( DFAgentDescription descs[] )
      {
         // Check we've got some results - if not just return
         if ( descs == null || descs.length <= 0 )
         {
            _matchmaker = null;
            DIAGNOSTICS.println("Did not find a matchmaker", DIAGNOSTICS.LEVEL_3);
         }
         else 
         {
            _matchmaker = descs[ 0 ].getAgentID();
            DIAGNOSTICS.println("Found a matchmaker: " + _matchmaker.getName(), this, DIAGNOSTICS.LEVEL_3);
            ACL reg = new ACL(); 
            reg.setOntology(IHCONSTANTS.IHELP_ONT);
            reg.setReceiverAID( getProxyAID()); 
            reg.addReplyToAID( _matchmaker);
            reg.setSenderAID( getAID());
            reg.setPerformative( FIPACONSTANTS.REQUEST);
            reg.setContentObject( IHCONSTANTS.REGISTER + " " + getProxyAID().toString());
            reg.setProtocol( FIPACONSTANTS.FIPA_REQUEST );
            forward(reg);
         }

      } // DFSearchResults()


      public void handleAgree(Conversation c)
      {
      }

      public void handleInform(Conversation c)
      { // once get a message back from matchmaker, we are done. 
         done();
      }

   }

   /** Handles all incoming i-help requests
   */
   public class HandleRequestTask extends Task
   {


      private ACL _acl;
      private Conversation _conv;
      

      public HandleRequestTask ( ACL acl, Conversation conv) 
      {
         _acl = acl;
         _conv = conv;
      }
      

      public void startTask() 
      {

         String s = (String)_acl.getContentObject();
         StringTokenizer st = new StringTokenizer(s);
         int n = st.countTokens();
         String function = st.nextToken(); 
         if (function.equals(IHCONSTANTS.KEYWORD_ADD))
         {
            if (n == 2) 
            {
               addKeyword(st.nextToken()); 
               sendReply(_acl, FIPACONSTANTS.AGREE);
               sendReply(_acl, FIPACONSTANTS.INFORM, getKeywords() );
               done(IHCONSTANTS.KEYWORD_ADD);
            }
         }
         else if (function.equals(IHCONSTANTS.KEYWORD_LIST))
         {
            if (n == 1) 
            {
               sendReply(_acl, FIPACONSTANTS.AGREE); 
               sendReply(_acl, FIPACONSTANTS.INFORM, getKeywords() );
               done(IHCONSTANTS.KEYWORD_LIST);
            }
         }
         else if (function.equals(IHCONSTANTS.REGISTERUI))
         {
            // set the UIAgent
            if (n == 2) 
            {
               String onoff = st.nextToken(); 
               if ( onoff.equals(IHCONSTANTS.ONLINE) ) 
               { 
                  setUIAgent(_acl.getSenderAID());
                  setStatus(true);
               }
               else if ( onoff.equals(IHCONSTANTS.OFFLINE) ) 
               {
                  setUIAgent(null);
                  setStatus(false);
               }
               sendReply(_acl, FIPACONSTANTS.AGREE); 
               sendReply(_acl, FIPACONSTANTS.INFORM, getStrStatus() );
               done(IHCONSTANTS.REGISTERUI);
            }
         }
         else if (function.equals(IHCONSTANTS.HELP))
         {
            // someone needs help
            if (n == 3) 
            {
               sendReply(_acl, FIPACONSTANTS.AGREE); 
               String person = st.nextToken();
               String kw = st.nextToken();
               
               if ( _status && (hasKeyword(kw)))
                  sendReply(_acl, FIPACONSTANTS.INFORM,  IHCONSTANTS.YES);
               else 
                  sendReply(_acl, FIPACONSTANTS.INFORM,  IHCONSTANTS.NO);
               done(IHCONSTANTS.HELP);
            }
         }
         else if (function.equals(IHCONSTANTS.FIND))
         {
            // find a helper
            if (n == 3) 
            {
               sendReply(_acl, FIPACONSTANTS.AGREE); 
               String person = st.nextToken();
               String kw = st.nextToken();
               newTask( new FindHelpTask( kw, person ), IHCONSTANTS.FIND_TIMEOUT); 
            }
         } 
         else 
         {
            sendNotUnderstood(_acl); 
            done("Not Understood");
         } 
      }


      public void donePersAgent_FindHelpTask(Object o) 
      {
         String helpers = (String) o;
         sendReply(_acl, FIPACONSTANTS.INFORM, helpers);
         done(IHCONSTANTS.FIND);
      } 


      public void timeoutPersAgent_FindHelpTask(Task t) 
      {
         // send message back saying could not find any helpers! 
         sendReply(_acl, FIPACONSTANTS.INFORM, "No HELPers Found");
         done(IHCONSTANTS.FIND);
      } 

   } // HandleRequestTask


   /** Find a helper. Send a message to the MM and wait for a reply
   */
   public class FindHelpTask extends Task 
   {

      private String _kw; 
      private String _person; 


      public FindHelpTask(String kw, String person) 
      {
         _kw = kw; 
         _person = person;
      }


      protected void startTask()
      {
         ACL req = new ACL();
         req.setSenderAID( getAID());
         // send message to proxy and add real addressee to replyto
         req.setReceiverAID( getProxyAID()); 
         req.addReplyToAID(_matchmaker); 
         //
         req.setPerformative( FIPACONSTANTS.REQUEST );
         req.setOntology( IHCONSTANTS.IHELP_ONT );
         req.setContentObject( IHCONSTANTS.FIND + " " +_person + " " + _kw);
         req.setProtocol(FIPACONSTANTS.FIPA_REQUEST);
         forward( req );
      }

      public void handleAgree(Conversation c)
      {
      }


      public void handleInform(Conversation c)
      {
         ACL reply = c.getACL( c.getLatestMessageIndex());
         String helpers = (String) reply.getContentObject();
         done(helpers);
      }

   } // FindHelpTask


   public static void main( String args[] ) {

      if ( args == null || args.length < 2 ) {

         System.out.println( "Usage: java fipaos.tutorial.PersAgent <platform.profile> <name>" );
         return;
      }

      new PersAgent( args[0], args[1], "FIPA-OS", null, "REPLICANT", "NONE" );

   } // main()

} // PersAgent
