/* 
   IHELP Simulation - UIAgent V1.0 

   Written by: 
      Alan Fedoruk                                           -- amf673@mail.usask.ca
      MADMUC Lab                                             -- http://bistrica.usask.ca/madmuc/index.htm
      Dept. of Computer Science, University of Saskatchewan. -- http://www.cs.usask.ca

   Copyright (c) 2001 Alan M. Fedoruk
*/

package ihelp;

// Import the fipa-os classes
import fipaos.agent.*;
import fipaos.agent.conversation.*;
import fipaos.agent.task.*;
import fipaos.mts.*;
import fipaos.ont.fipa.*;
import fipaos.ont.fipa.fipaman.*;
import fipaos.parser.acl.ACLMessage;

// Import registration exception classes
import fipaos.platform.ams.AMSRegistrationException;
import fipaos.platform.df.DFRegistrationException;

import fipaos.util.DIAGNOSTICS;

import repserver.*;

import java.util.*;
import java.io.*;


/** A user interface agent. This is a subclass of the standard FIPAOSAgent rather 
    than a replicant. Since this is a user interface, there is no need to replicate 
    this. Shows that regular agents and replicate agents can co-exist. 
*/
public class UIAgent extends FIPAOSAgent 
{

   /* This Agent type */
   private static final String AGENT_TYPE = IHCONSTANTS.UIAGENT; 

   /*  GUI for this UIAgent */
   private UIAgentGUI _gui;

   /* The AgentID  of my personal agent */
   private AgentID _persagent;
  
   /* my name */
   private String _persname;

   /* my status, online or offline */
   private String _status; 

   /* the keywords that I know about */
   private String _keywords; 

   /* used for timing requests */
   private long _starttime;

   /* used for timing requests */
   private long _endtime;

   /* used for logging timing data */
   private PrintWriter _logfile;  


   public UIAgent( String platform, String name, String ownership, String persname) 
   {
   
      super( platform, name, ownership);
      _persname = persname;
      _status = IHCONSTANTS.OFFLINE;
      _keywords = "";

      try 
      {
         _logfile = new PrintWriter( new BufferedWriter(new FileWriter(name + ".log")));
      }
      catch ( Exception e)
      {
         System.out.println(e);
      }
    
      // Create a GUI for this Agent so we can see what is going on.
      try 
      {
         _gui = new UIAgentGUI( getAID().getName(), this );
      } 
      catch ( Throwable t ) 
      { 
         System.out.println(t);
      }
    
      // Instantiate a new IdleTask to receieve all incoming 
      // communications which are not associated with any-other Task

      setListenerTask( new IdleTask() );
      
      // Register with the local AMS (i.e. the platform)
      try 
      {
         registerWithAMS();
         DIAGNOSTICS.println( "Registered with AMS", this, DIAGNOSTICS.LEVEL_MAX );
      } catch (AMSRegistrationException amsre) 
      {
         DIAGNOSTICS.println(amsre, this, DIAGNOSTICS.LEVEL_MAX );
         String reason = amsre.getExceptionReason();
      
         if ( reason == null || !reason.equals( FIPAMANCONSTANTS.AGENT_ALREADY_REGISTERED)) 
         {
            shutdown();
            return;
         }
      } // end catch 

      try 
      {
         registerWithDF( AGENT_TYPE );
         DIAGNOSTICS.println("Registered with DF", this, DIAGNOSTICS.LEVEL_MAX );
      } catch (DFRegistrationException dfre) 
      {
         DIAGNOSTICS.println(dfre, this, DIAGNOSTICS.LEVEL_MAX );
      
         String reason = dfre.getExceptionReason();
      
         if ( reason == null || !reason.equals( FIPAMANCONSTANTS.AGENT_ALREADY_REGISTERED) ) 
         {
            shutdown();
            return;
         }
      } // end catch

  
   } // UIAgent()
  

   /**
    * Invoked when this Agent is about to be shutdown - this overrides the
    * shutdown() method in the FIPAOSAgent class to enable the CountAgentGUI
    * associated with this CountAgent to be removed
    */
   public synchronized void shutdown() 
   {
      if (registeredWithDF() == true) 
      {
         try 
         {
            deregisterWithDF();
            DIAGNOSTICS.println("Deregistered with DF", this, 
                  DIAGNOSTICS.LEVEL_MAX );
         } catch (DFRegistrationException dfre) 
         {
            // A problem deregistering occured..... we can obtain 
            // the reason though!
            String reason = dfre.getExceptionReason();

            DIAGNOSTICS.println(dfre + ":" + reason, this, 
                  DIAGNOSTICS.LEVEL_MAX ); 
         } // end catch
      } // if 
    
      if (registeredWithAMS() == true) 
      {
         try 
         {
            deregisterWithAMS();
            DIAGNOSTICS.println("Deregistered with AMS", this, 
                  DIAGNOSTICS.LEVEL_MAX ); 
         } catch (AMSRegistrationException amsre) 
         {
            String reason = amsre.getExceptionReason();
     
            DIAGNOSTICS.println(amsre + ":" + reason, this, 
                  DIAGNOSTICS.LEVEL_MAX );
         } // catch
      } // if 
     
      // Dispose of the GUI

      if ( _gui != null ) 
      {
         _gui.dispose();
      }

      _logfile.flush();
      _logfile.close();
    
      // Now call shutdown in the agent shell to release the core components
      super.shutdown();

   } // shutdown()
  

   /**
    * This method is invoked when GUI should be shown - generally
    * this is invoked by the AgentLoader when the Agent is in the List
    * of running Agents - the default implementation in FIPAOSAgent does 
    * nothing.
    */

   public void activate() 
   {
      // Check if we've got a GUI, if so, show it
      if( _gui != null ) 
      {
         _gui.setVisible( true );
      }

   } // activate()

    
   /** start a FindPersAgentTask to find my personal agent
   */
   public void findPersAgent() 
   {
      _gui.write("Started Search for Personal Agent...\n");
      _tm.newTask( new FindPersAgentTask(), 30000 );
   }


   /** once FindPersAgentTask completes update the GUI
   */
   public void doneUIAgent_FindPersAgentTask(Task t)
   {
      _gui.updateGUI();
   }


   /** if the FindPersAgentTask timeouts 
   */
   public void timeoutUIAgent_FindPersAgentTask(Task t)
   {
      _gui.write("Could not contact the Personal Agent\n");
      _gui.updateGUI();
   }

 
   /** if the personal agent is online, start the 
       AddKeywordTask task
   */
   public void addKeyword( String kw)
   {
      if ( _persagent == null) 
      { 
         _gui.write("Must Find Personal Agent First!\n"); 
         return;
      }
      else
         _tm.newTask( new AddKeywordTask( kw ) );
   }


   /** once the AddKeywordTask is done.
   */
   public void doneUIAgent_AddKeywordTask(Task t)
   {
      _gui.updateGUI();
   }


   public void switchStatus() 
   {
      if (_persagent == null) 
      { 
         _gui.write("Must Find Personal Agent First!\n"); 
         return;
      }
      else
      { 
         _tm.newTask( new SwitchStatusTask() );
      }
   }


   public void doneUIAgent_SwitchStatusTask(Task t)
   {
      _gui.updateGUI();
   }


   /** start a FindHelpTask to find helpers for the given keyword
   */
   public void findHelp( String kw)
   {
      if (_persagent == null) 
      { 
         _gui.write("Must Find Personal Agent First!\n"); 
         return;
      }
      else
      {
         _starttime = System.currentTimeMillis();
         _tm.newTask( new FindHelpTask(kw), IHCONSTANTS.FIND_TIMEOUT );
      }
      
   }

   /** Get the number of current conversations */
   public int getConvs()
   {
      return getCM().getConvCount();
   }



   public void printCounts()
   {
      System.out.println("UIAgent: " + getAID().getName() + " Tasks = " + _tm.getTaskCount());
      System.out.println("UIAgent: " + getAID().getName() + " Convs = " + getConvs());
   }

   /** when the FindHelpTask is done... 
   */
   public void doneUIAgent_FindHelpTask(Object o)
   {
      _endtime = System.currentTimeMillis();
      long etime = _endtime - _starttime; 
      _gui.write("Helpers: " + (String)o + " " + etime + "ms.\n"); 
      _gui.updateGUI();
      _logfile.println( "Find: Success " + etime );
   }

   /** when the FindHelpTask times out
   */
   public void timeoutUIAgent_FindHelpTask(Task t)
   {
      _endtime = System.currentTimeMillis();
      long etime = _endtime - _starttime; 
      _gui.write("Unable to find a helper in: " + etime + "ms.\n"); 
      _gui.updateGUI();
      _logfile.println( "Find: Failure" + etime );
   }


   public String getStatus() 
   {
      return _status;
   }


   public String getKeywords()
   {
      return _keywords;
   }


   public String getPersAgent()
   {
      if ( _persagent == null ) 
         return ""; 
      else 
         return _persagent.getName();
   }


   public class IdleTask extends Task 
   {


      public IdleTask() 
      {
      } // IdleTask()
    
    
      protected void startTask() 
      {
      } // StartTask()

   } // IdleTask()



   public class AddKeywordTask extends Task 
   {

      String _kw = "";


      public AddKeywordTask(String kw) 
      {
         _kw = kw;
      }


      protected void startTask() 
      {
         ACL add = new ACL() ;
         add.setPerformative( FIPACONSTANTS.REQUEST);
         add.setSenderAID( getAID());
         add.setReceiverAID( _persagent);
         add.setOntology( IHCONSTANTS.IHELP_ONT);
         add.setContentObject( IHCONSTANTS.KEYWORD_ADD + " " + _kw); 
         add.setProtocol( FIPACONSTANTS.FIPA_REQUEST);
         forward( add );
      }


      public void handleAgree( Conversation c)
      {
      }


      public void handleInform(Conversation c) 
      {
         _gui.write("Keyword " + _kw + " Added.\n");
         ACL inform = c.getACL( c.getLatestMessageIndex() );
         _keywords = (String) inform.getContentObject();
         done();
      }
   } // AddKeyWordTask


   public class GetKeywordsTask extends Task 
   {

      String _kw = "";


      public GetKeywordsTask() 
      {
      }


      protected void startTask() 
      {
         ACL list = new ACL() ;
         list.setPerformative( FIPACONSTANTS.REQUEST);
         list.setSenderAID( getAID());
         list.setReceiverAID( _persagent);
         list.setOntology( IHCONSTANTS.IHELP_ONT);
         list.setContentObject( IHCONSTANTS.KEYWORD_LIST ); 
         list.setProtocol(FIPACONSTANTS.FIPA_REQUEST);
         forward( list );
      }

      public void handleAgree(Conversation c) 
      {
      }

 
      public void handleInform(Conversation c) 
      {
         ACL inform = c.getACL( c.getLatestMessageIndex());
         String kw = (String) inform.getContentObject(); 
         done(kw);
      }

   } // AddKeyWordTask




   public class FindHelpTask extends Task 
   {

      String _kw = "";


      public FindHelpTask(String kw) 
      {
         _kw = kw;
      }


      protected void startTask() 
      {
         ACL find = new ACL() ;
         find.setPerformative( FIPACONSTANTS.REQUEST);
         find.setSenderAID( getAID());
         find.setReceiverAID( _persagent);
         find.setOntology( IHCONSTANTS.IHELP_ONT);
         find.setContentObject( IHCONSTANTS.FIND + " " + getPersAgent() + " " + _kw); 
         find.setProtocol(FIPACONSTANTS.FIPA_REQUEST);
         forward( find );
      }


      public void handleAgree(Conversation c) 
      {
      }

 
      public void handleInform(Conversation c) 
      {
         ACL inform = c.getACL( c.getLatestMessageIndex());
         String helpers = (String) inform.getContentObject(); 
         done(helpers);
      }

   } // AddKeyWordTask


   /** Send a message to my personal agent to 
       toggle my status. 
   */
   public class SwitchStatusTask extends Task
   {


      public SwitchStatusTask() 
      {
      }


      protected void startTask()
      {
         ACL reg = new ACL() ;
         reg.setPerformative( FIPACONSTANTS.REQUEST);
         reg.setSenderAID( getAID());
         reg.setReceiverAID( _persagent);
         reg.setOntology( IHCONSTANTS.IHELP_ONT);
         if ( _status.equals(IHCONSTANTS.OFFLINE)) 
            reg.setContentObject( IHCONSTANTS.REGISTERUI + " " + IHCONSTANTS.ONLINE); 
         else 
            reg.setContentObject( IHCONSTANTS.REGISTERUI + " " + IHCONSTANTS.OFFLINE); 
         reg.setProtocol(FIPACONSTANTS.FIPA_REQUEST);
         forward(reg);
      }


      public void handleAgree(Conversation c)
      {
      }


      public void handleInform(Conversation c) 
      {
         if ( _status.equals(IHCONSTANTS.OFFLINE)) 
            _status = IHCONSTANTS.ONLINE; 
         else 
            _status = IHCONSTANTS.OFFLINE; 
         done();
      }

   }


   /** Find my personal agent from the DF, then contact it. 
   */
   public class FindPersAgentTask extends Task 
   {


      public FindPersAgentTask()
      {
      } 


      protected void startTask() 
      {
         // Create the DFAgentDescription describing the attributes we wish
         // to constrain in our search for DataAgents
         DFAgentDescription df_desc = new DFAgentDescription();
      
         ServiceDescription sd = new ServiceDescription();
         sd.setServiceName( _persname);
         sd.setServiceType( _persname); 
         df_desc.setAgentServices( new ServiceDescription[] { sd } );
      
         // Start a search for other Agents
         searchDF( df_desc );
   
      } // StartTask()


      public void DFSearchResults( DFAgentDescription descs[] ) 
      {
         // Check we've got some results - if not just return

         if ( descs == null || descs.length <= 0 ) 
         {
            _gui.write("Could not find Personal Agent!!!\n");
            return;
         }
        
         _persagent = descs[0].getAgentID();
         _gui.write( "Found " + _persagent.getName() + " my personal agent!\n" );

         // send a registerUI to the persagent to let it know I am here! 
         
         ACL reg = new ACL() ;
         reg.setPerformative( FIPACONSTANTS.REQUEST);
         reg.setSenderAID( getAID());
         reg.setReceiverAID( _persagent);
         reg.setOntology( IHCONSTANTS.IHELP_ONT);
         reg.setContentObject( IHCONSTANTS.REGISTERUI + " " + IHCONSTANTS.ONLINE); 
         reg.setProtocol( FIPACONSTANTS.FIPA_REQUEST);
         forward(reg);
      
      } // DFSearchResults()


      public void handleAgree(Conversation c)
      {
      }

      public void handleInform(Conversation c)
      {
         newTask ( new GetKeywordsTask() );
         _status = IHCONSTANTS.ONLINE;
      }


      public void doneUIAgent_GetKeywordsTask( Object o) 
      {
         _keywords = (String) o;
         done();
      }


      public void timeoutUIAgent_GetKeywordsTask( Task t) 
      {
         _gui.write("Could Not Get Keywords\n");
         done();
      }

   } // FindPersAgentTask


   /**
   * Test-harness
   */
   public static void main( String args[] ) 
   {

      if ( args == null || args.length < 3 ) 
      {
         System.out.println( "Usage: java fipaos.tutorial.UIAgent <platform.profile> <name> <person>" );
         return;
      }

      new UIAgent( args[0], args[1], "FIPA-OS", args[2]);

   } // main()

} // UIAgent
