/* 
   IHELP Simulation - MatchMaker Agent V1.0 

   Written by: 
      Alan Fedoruk                                           -- amf673@mail.usask.ca
      MADMUC Lab                                             -- http://bistrica.usask.ca/madmuc/index.htm
      Dept. of Computer Science, University of Saskatchewan. -- http://www.cs.usask.ca

   Copyright (c) 2001 Alan M. Fedoruk
*/

package ihelpnorep;

import fipaos.agent.*;
import fipaos.agent.conversation.*;
import fipaos.agent.task.*;
import fipaos.mts.*;
import fipaos.ont.fipa.*;
import fipaos.ont.fipa.fipaman.*;
import fipaos.parser.*;
import fipaos.parser.acl.ACLMessage;
import fipaos.util.DIAGNOSTICS;
import fipaos.platform.*;
import fipaos.platform.ams.*;
import fipaos.platform.df.*;

import java.util.*;
import java.io.*;
import ihelp.*;

/** MMAgent is the IHelp matchmaker agent

    It maintains a hashtable of personal agents, broadcasts help requests 
    to all the PersAgents that it knows about and assembles the replies 
    that it gets.  */
public class MMAgent extends FIPAOSAgent
{

   /* my type */
   private static final String AGENT_TYPE = IHCONSTANTS.MMAGENT;

   /* Hashtable of the agents that I know about. */
   private Hashtable _agents; 

   /* A random number generator for simulated failures */
   private Random _r = new Random();


   public MMAgent( String platform, String name, String ownership ) 
   {
   
      super( platform, name, ownership );

      _agents = new Hashtable();

      setListenerTask( new IdleTask());

  // Register with the local AMS (i.e. the platform)
      try 
      {
         registerWithAMS();
         DIAGNOSTICS.println( "Registered with AMS", this, DIAGNOSTICS.LEVEL_MAX );
      } catch (AMSRegistrationException amsre) 
      {
         DIAGNOSTICS.println(amsre, this, DIAGNOSTICS.LEVEL_MAX );
         String reason = amsre.getExceptionReason();
      
         if ( reason == null || !reason.equals( FIPAMANCONSTANTS.AGENT_ALREADY_REGISTERED)) 
         {
            shutdown();
            return;
         }
      } // end catch 
    
      // Register with the local DF
      try 
      {
         registerWithDF( AGENT_TYPE );
         DIAGNOSTICS.println("Registered with DF", this, DIAGNOSTICS.LEVEL_MAX );
      } catch (DFRegistrationException dfre) 
      {
         DIAGNOSTICS.println(dfre, this, DIAGNOSTICS.LEVEL_MAX );
      
         String reason = dfre.getExceptionReason();
      
         if ( reason == null || !reason.equals( FIPAMANCONSTANTS.AGENT_ALREADY_REGISTERED) ) 
         {
            shutdown();
            return;
         }
      } // end catch



   } // MMAgent()
  


   public synchronized void shutdown() 
   {
      if (registeredWithDF() == true) 
      {
         // Attempt to deregister with DF
         try 
         {
            deregisterWithDF();
            DIAGNOSTICS.println("Deregistered with DF", this, DIAGNOSTICS.LEVEL_MAX );
         } catch (DFRegistrationException dfre) 
         {
            // A problem deregistering occured..
            String reason = dfre.getExceptionReason();
            DIAGNOSTICS.println(dfre + ":" + reason, this, DIAGNOSTICS.LEVEL_MAX ); 
         } // end catch
      } // if 
    
      if (registeredWithAMS() == true) 
      {
         // Attempt to deregister with AMS
         try 
         {
            deregisterWithAMS();
            DIAGNOSTICS.println("Deregistered with AMS", this, DIAGNOSTICS.LEVEL_MAX ); 

         } catch (AMSRegistrationException amsre) 
         {
            // A problem deregistering occured.
            String reason = amsre.getExceptionReason();
            DIAGNOSTICS.println(amsre + ":" + reason, this, DIAGNOSTICS.LEVEL_MAX );
         } // catch
      } // if 

      super.shutdown();
   } // shutdown()

    

   /** true if the given AgentID is registered with this MM
   */
   public boolean hasPersAgent( AgentID aid) 
   {
      if (_agents.containsKey( aid.getName()))
         return true; 
      else 
         return false; 
   }


   /** Add the given AgentID to the MM
   */
   public void addPersAgent( AgentID aid) 
   {
      delPersAgent(aid); 
      _agents.put( aid.getName(), aid);
   }


   /** Remove the given AgentID from the MM
   */
   public void delPersAgent( AgentID aid) 
   {
      if (_agents.containsKey( aid.getName()))
         _agents.remove(aid.getName());
   }


   /** List, as a string the names of all the agents 
       currently registered with this matchmaker
   */
   public String listAgents() 
   {
      String res = "";
      for (Enumeration e = _agents.keys(); e.hasMoreElements(); )
      {
         res = res + " " + e.nextElement();
      }
      return res;
   }


   /** generate a random number, and if the number is in the 
       range to fail, shutdown this agent. 
   */
   public void checkForFailure()
   {
      float r = _r.nextFloat();
      if ( r <= IHCONSTANTS.FAIL_RATE) 
      {
         System.out.println(getAID().getName() + " is failing ");
         shutdown();
      }
   }


   /** Send a reply message in response to the given ACL message
       @param acl The ACL object to reply to
       @param reason Any additional message
   */
   public boolean sendReply( ACL orig_acl, String performative, String content )
   {
      try
      {
         // Duplicate ACL
         ACL acl = (ACL) orig_acl.clone();
            
         // Setup message
         acl.setPerformative( performative );
         acl.setReceiverAID( acl.getSenderAID() );
         acl.setSenderAID( getAID() );
         acl.setInReplyTo( acl.getReplyWith() );
         acl.setReplyWith( null );
            
         // Set content
         acl.setContentObject( content );
            
         // Attempt to send the reply
         forward( acl );
            
         // We're done
         return true;
      }
      catch( Throwable t )
      {
         DIAGNOSTICS.println( "Error occured when sending reply", t, this,
         DIAGNOSTICS.LEVEL_MAX );
         return false;
      }
   } // sendReply()

   
   public boolean sendReply( ACL orig_acl, String performative)
   {
      return sendReply(orig_acl, performative, ""); 
   }


   /** Get the number of current conversations */
   public int getConvs()
   {
      return getCM().getConvCount();
   }


   /** remove the given conversation */
   public void removeConv(Conversation conv)
   {
      try 
      {
         getCM().removeConversation(conv.getConversationID());
      }
      catch (Exception e)
      {
         System.out.println(e);
      }
   }


   /** IdleTask 
   */
   public class IdleTask extends Task 
   {

      public IdleTask() 
      {
      } // IdleTask()
    
    
      protected void startTask() 
      {
      } // StartTask()


      public void handleRequest( Conversation conv ) 
      {
         // Get the type of request
         ACL acl = conv.getACL( conv.getLatestMessageIndex() );

         if ( acl.getOntology().equalsIgnoreCase( IHCONSTANTS.IHELP_ONT ) ) 
         {
            newTask ( new HandleRequestTask(conv, acl), conv); 
         } // if
         else 
            sendNotUnderstood(acl);

      } // handleRequest()


      public void doneMMAgent_HandleRequestTask(Object o)
      {
         String function = (String) o;
         // after every request
         // checkForFailure();
         // if the function was something that changed my state, update my state
      }
    
   } // IdleTask()


   /** Handles all incoming i-help requests
   */
   public class HandleRequestTask extends Task
   {


      private ACL _acl;
      private Conversation _conv;
      // The operation performed
      private String _function; 
      

      public HandleRequestTask ( Conversation conv, ACL acl) 
      {
         _acl = acl;
         _conv = conv;
      }
      

      public void startTask() 
      {

         String content = (String)_acl.getContentObject();
         StringTokenizer st = new StringTokenizer(content);
         int n = st.countTokens();
         _function = st.nextToken(); 

         // register the sender personal agent with this MM. 
         if (_function.equals(IHCONSTANTS.REGISTER))
         {
            String aidString = 
                  content.substring( content.indexOf(IHCONSTANTS.REGISTER) + IHCONSTANTS.REGISTER.length());
            sendReply(_acl, FIPACONSTANTS.AGREE);
            try 
            {
               AgentID aid = new AgentID( aidString);
               addPersAgent(aid); 
               sendReply(_acl, FIPACONSTANTS.INFORM);
               done(_function);
            }
            catch (ParserException p)
            {
               sendReply(_acl, FIPACONSTANTS.FAILURE, "Malformed AgentID");
               done(_function);
            }
         }
         // deregister the sender personal agent with this MM. 
         else if (_function.equals(IHCONSTANTS.DEREGISTER))
         {
            if (n == 1) 
            {
               sendReply(_acl, FIPACONSTANTS.AGREE);
               AgentID aid = _acl.getSenderAID();
               if ( hasPersAgent(aid)) 
               {
                  delPersAgent(aid); 
               }
               sendReply(_acl, FIPACONSTANTS.INFORM);
            }
         }
         // find helpers capable of the keyword given. 
         else if (_function.equals(IHCONSTANTS.FIND))
         {
            // find helpers capable of the keyword given. 
            // later can add capability for more keywords.
            if (n == 3) 
            {
               sendReply(_acl, FIPACONSTANTS.AGREE);
               String person = st.nextToken();
               String kw = st.nextToken();
               newTask ( new FindTask(kw, person), IHCONSTANTS.FIND_TIMEOUT); 
            }
         }
         else 
         {
            sendNotUnderstood(_acl); 
            done(_function);
         } 
      }


      public void doneMMAgent_FindTask( Object o) 
      {
         sendReply(_acl, FIPACONSTANTS.INFORM, (String) o);
         done(_function);
      // send inform with results of search. 
      }


      public void timeoutMMAgent_FindTask( Task t) 
      {
         sendReply(_acl, FIPACONSTANTS.INFORM, "");
         done(_function);
      }


   } // HandleRequestTask


   /** FindTask sends a help request to all the personal agents that 
       this MM knows about. The responses are tabulated and sent 
       back to the personal agent that asked for help. 
   */
   public class FindTask extends Task 
   {

      /** count of how many agents I have to send to. 
      */
      private int _count; 
      private String _response = "";
      private String _keyword; 
      private String _person; 


      public FindTask (String kw, String person)
      {
         _keyword = kw;
         _person = person;
      } // FindTask()


      protected void startTask() 
      {
         _count = _agents.size();
         for (Enumeration e = _agents.keys(); e.hasMoreElements() ;) 
         {
            AgentID aid = (AgentID)_agents.get((String)e.nextElement());
            if ( !aid.getName().equals(_person) ) 
            { // dont ask myself!
               newTask( new SendFindTask( aid, _keyword, _person), IHCONSTANTS.FIND_TIMEOUT); 
            }
            else 
               _count--;
         }
      } // startTask()


      public void doneMMAgent_SendFindTask( Task t)
      {
         if ( t instanceof SendFindTask )
         {
            SendFindTask s = (SendFindTask) t;
            String response = s.getResponse(); 
            String aname = s.getAgent().getName();
            if ( response.equals("YES")) 
            { 
               _response = _response + " " + aname;
            }
         }
         decrementCount();
      }


      public void timeoutMMAgent_SendFindTask( Task t)
      {
         decrementCount();
      }


      public void errorMMAgent_SendFindTask( Task t, Throwable th)
      {
         decrementCount();
      } 


      private synchronized void decrementCount()
      {
         _count--; 
         if ( _count <= 0)
         {
            done(_response);
         }

      } // decrementCount()


   } // FindTask



   /** send a message to the given agent to see if the 
       agents user is willing and able to help. 
   */
   public class SendFindTask extends Task
   {


      /* the agent to send the request to */
      private AgentID _aid; 

      /* the keyword to ask for */
      private String _keyword; 

      /* the person asking for help */
      private String _person; 
      
      /* the response from this agent. Set to NO by default */
      private String _response = IHCONSTANTS.NO; 


      /** return the AgentID of the agent we are asking 
      */
      public AgentID getAgent() 
      {
         return _aid; 
      }


      /** return the response that we got. 
      */
      public String getResponse()
      {
         return _response;
      } 


      public SendFindTask(AgentID aid, String kw, String person) 
      {
         _aid = aid; 
         _keyword = kw;
         _person = person;
      }


      protected void startTask()
      {
         // send the request. Since we are initiated this conversation 
         // and this is a replicated agent... we need to do a little extra  
         ACL acl = new ACL();
         acl.setPerformative( FIPACONSTANTS.REQUEST );
         acl.setSenderAID( _owner.getAID() );
         // send the message to the group_proxy
         acl.setReceiverAID( _aid);
         // add the agent to forward to in the replyto
         acl.setContentObject( IHCONSTANTS.HELP + " " + _person + " " + _keyword );
         acl.setOntology( IHCONSTANTS.IHELP_ONT );
         acl.setProtocol( FIPACONSTANTS.FIPA_REQUEST );
         forward( acl );
      }


      public void handleAgree(Conversation conv) 
      {
         // ignore the agree
      } 


      public void handleInform(Conversation conv) 
      {
         ACL acl = conv.getACL( conv.getLatestMessageIndex() );
         _response = (String) acl.getContentObject(); 
         done();
      } 

   } // SendFindTask


   /** Test harness */
   public static void main( String args[] ) {

      if ( args == null || args.length < 2 ) {

         System.out.println( "Usage: java fipaos.tutorial.MMAgent <platform.profile> <name>" );
         return;
      }

      new MMAgent( args[0], args[1], "FIPA-OS" );

   } // main()

} // MMAgent
