/* 
   Replication Server - RepGroup V1.0, November 6, 2001 

   Written by: 
      Alan Fedoruk                                           -- amf673@mail.usask.ca
      MADMUC Lab                                             -- http://bistrica.usask.ca/madmuc/index.htm
      Dept. of Computer Science, University of Saskatchewan. -- http://www.cs.usask.ca

   Copyright (c) 2001 Alan M. Fedoruk
*/

package repserver;

// fipa imports
import fipaos.ont.fipa.fipaman.*;
import fipaos.ont.fipa.*;

// java imports
import java.io.*;
import java.util.*;


/** This class represents a Replicant Group
    A replicant group consists of a list of RepAgentIDs,
    the latest serialized state, a pointer to an active agent,
    and a message proxy (RepProxy). 
*/
public class RepGroup implements Serializable, Cloneable 
{

   /** Name of this replicant group. */
   private String _name;

   /** A linked list of agent ids in this replicant group */
   private Hashtable _replicants;

   /** The agent id of the active agent */
   private RepAgentID _active_agent; 

   /** The previous states of the active agent */
   private Vector _state;
	
   /** The proxy agent for this group */
   private RepProxy _repproxy; 

   /** The platform */
   private String _platform;

   private String _agent_type;

   /** the type of replication */

   private String _replicationtype;

   /** Constructs a blank RepGroup.  
   */
   public RepGroup() 
   {
     _name = null;
     _replicants = new Hashtable();
     _active_agent = null; 
     _state = new Vector(); 
     setReplicationType(RSCONSTANTS.HOT_STANDBY);

   } // RepGroup()
	

   /** Constructs a RepGroup with the given name 
       @param n string name to name this repgroup 
   */
   public RepGroup(String platform, String n, String type ) 
   {
      // Invoke default constructor
      this();
      _name = n; 
      _platform = platform; 
      _agent_type = type;

      // start a RepProxy agent. This will handle incoming communications. 
      _repproxy = new RepProxy(_platform, _name,  FIPACONSTANTS.FIPA_OS, _agent_type, this);

   } // RepGroup()
  
	
   public void setReplicationType(String reptype)
   {
      _replicationtype = reptype;
   }
   public String getReplicationType()
   {
      return _replicationtype;
   }
	
   /** Sets the name of this RepGroup
       @param name The name to set
   */
   public void setName( String name ) 
   {
      _name = name;
   }


   /** Returns the name of this repgroup 
   */
   public String getName()
   {
      return _name;
   }

   /** Add an agent to this replicant group. 
       @param aid - the agent to add
   */
   public void addAgent( RepAgentID aid) 
   {
      if ( ! _replicants.contains(aid) ) 
      {
         _replicants.put( aid.getName(), aid);
      }

   } // addAgent()


   /** Delete an agent from this replicant group. 
       @param aid - the agent to delete
   */
   public void delAgent( RepAgentID aid) 
   {
      _replicants.remove( aid.getName());
   }

 
   /** get the Active Agent 
   */
   public RepAgentID getActive() 
   {
      return _active_agent; 
   } 


   /** Set the active agent to be the first agent found 
       alive Return true if one is found, false otherwise.  
   */
   public boolean setActive() 
   {
      Enumeration e = _replicants.keys();
      while ( e.hasMoreElements() ) 
      {
         RepAgentID r = (RepAgentID) _replicants.get((String)e.nextElement());
         if ( r.getPing() ) 
         {
            setActive(r); 
            return true; 
         }
      }
      return false; 

   } // setActive()


   /** Set the active agent to be the agent named by aid. 
       Return true if aid is found and is alive, false otherwise.  
   */
   public boolean setActive(RepAgentID aid) 
   {
      _active_agent = aid; 
      return true; 
   }

   /** Get the message proxy 
   */
   public RepProxy getProxy() 
   {
      return _repproxy;
   }

   /** Get the agentID of the mail proxy 
   */
   public AgentID getProxyAID() 
   {
      return _repproxy.getAID();
   }


   /** convert this repgroup to a string. 
   */
   public String toString() 
   {
      String res = "";
      for ( Enumeration e = _replicants.keys(); e.hasMoreElements(); )
      {
         res = res + " " + ((RepAgentID)_replicants.get((String)e.nextElement())).getName(); 
      }
      return res;
   } 

   
   /** Return an array of the agents in this group. 
   */  
   public RepAgentID[] getAgents() 
   {
      int rs = _replicants.size();

      RepAgentID res[] = new RepAgentID[rs];

      int i = 0;
      for ( Enumeration e = _replicants.keys(); e.hasMoreElements(); )
      {
         res[i] = (RepAgentID) _replicants.get((String)e.nextElement());  
         i++;
      }
      return res;

   } // getAgents()


   /** Set the ping flag for the given agent. 
       @param raid the agent in question. 
       @param val what to set the flag to. 
   */
   public void setPing(RepAgentID raid, boolean val) 
   {
      ((RepAgentID)_replicants.get(raid.getName())).setPing(val);      
   }


   /** Return the ping flag for the given agent.  
       @param raid the agent in question.
   */
   public boolean getPing(RepAgentID raid)
   {
      return ((RepAgentID)_replicants.get(raid.getName())).getPing();      
   } 


   /** add the state to the stack.
       @param state a byte stream containing the agent state
   */
   public void setState(byte[] state) 
   {
      if ( _state.size() == RSCONSTANTS.STATES) 
      {
         _state.removeElementAt(0);
      }
      _state.addElement(state); 
   } 


   /** return the most recent value of the _state vector
   */
   public byte[] getState() 
   {
      if ( _state.isEmpty() )
         return null;
      else 
         return (byte[]) _state.lastElement();
   } // getState


   /** return the number of replicants in this replicant group 
   */
   public int countReps() 
   {
      return _replicants.size();
   }


   /** set the updatetime for the given agent. 
   */
   public void setUpdateTime(RepAgentID raid, Date d) 
   {
      ((RepAgentID)_replicants.get(raid.getName())).setUpdate(d);      
   }


   /** get the updatetime for the given agent. 
   */
   public Date getUpdateTime(RepAgentID raid) 
   {
      return ((RepAgentID)_replicants.get(raid.getName())).getUpdate();      
   } 


   /** remove any dead agents from the rep group. 
   */
   public void removeDead()
   {
      for (Enumeration e = _replicants.keys(); e.hasMoreElements();) 
      {
         RepAgentID raid = (RepAgentID)_replicants.get( (String)e.nextElement() );
         if ( !raid.getPing() ) 
            delAgent(  raid);
      }
   }

} // RepGroup
