/* 
   Replication Server - RepProxy V1.0, November 6, 2001 

   Written by: 
      Alan Fedoruk                                           -- amf673@mail.usask.ca
      MADMUC Lab                                             -- http://bistrica.usask.ca/madmuc/index.htm
      Dept. of Computer Science, University of Saskatchewan. -- http://www.cs.usask.ca

   Copyright (c) 2001 Alan M. Fedoruk
*/


package repserver;

// Import the fipa-os classes
import fipaos.agent.*;
import fipaos.agent.conversation.*;
import fipaos.agent.task.*;
import fipaos.mts.*;
import fipaos.ont.fipa.*;
import fipaos.ont.fipa.fipaman.*;
import fipaos.parser.acl.ACLMessage;
import fipaos.util.DynamicClass;

// Import registration exception classes
import fipaos.platform.ams.AMSRegistrationException;
import fipaos.platform.df.DFRegistrationException;
import fipaos.util.DIAGNOSTICS;

import java.util.*;

import localpostoffice.*;


/** RepProxy is the message proxy for a replicant group. 
    Each RepGroup has one. 

    RepProxy waits (IdleTask) for messages either from one of its replicant group 
    or from another agent in the MAS. 

    Messages from agents in the replicant group arrive as PROPAGATE performatives. 

*/
public class RepProxy extends FIPAOSAgent implements LocalMsg
{

   /** This Agent type */
   private static final String AGENT_TYPE = RSCONSTANTS.REPPROXY;

   /** the repgroup associated with this proxy */
   private RepGroup _repgroup; 

   /** Location of platform profile */
   private String _platform_profile;

   /** Agent type for DF registration */
   private String _agent_type; 

   private LocalPostOffice _lpo = null; 

   /** constructor - sets up an IdleTask as the default
       ListenerTask for incoming messages, and starts off registration
       with the AMS and DF
   */
   public RepProxy (String platform, String name, String ownership, String agent_type, RepGroup repgroup ) 
   {
   
      super( platform, name, ownership );
    
      _platform_profile = platform;
      _repgroup = repgroup; 
      _agent_type = agent_type;

      // Instantiate a new IdleTask to receieve all incoming 
      // communications which are not associated with any other Task

      setListenerTask( new IdleTask() );
  
      // Register with the local AMS (i.e. the platform)
      try 
      {
         registerWithAMS();
         DIAGNOSTICS.println( "Registered with AMS", this, DIAGNOSTICS.LEVEL_MAX );
      } catch (AMSRegistrationException amsre) 
      {
         DIAGNOSTICS.println(amsre, this, DIAGNOSTICS.LEVEL_MAX );
         String reason = amsre.getExceptionReason();
      
         if ( reason == null || !reason.equals( FIPAMANCONSTANTS.AGENT_ALREADY_REGISTERED)) 
         {
            shutdown();
            return;
         }
      } // end catch 
    
      // Register with the local DF
      try 
      {
         registerWithDF( _agent_type );
         DIAGNOSTICS.println("Registered with DF", this, DIAGNOSTICS.LEVEL_MAX );
      } catch (DFRegistrationException dfre) 
      {
         DIAGNOSTICS.println(dfre, this, DIAGNOSTICS.LEVEL_MAX );
      
         String reason = dfre.getExceptionReason();
      
         if ( reason == null || !reason.equals( FIPAMANCONSTANTS.AGENT_ALREADY_REGISTERED) ) 
         {
            shutdown();
            return;
         }
      } // end catch

    
   } // RepProxy()
  

   /** Invoked when this Agent is about to be shutdown - this overrides the
       shutdown() method in the FIPAOSAgent class to enable the RepServerGUI
       associated with this RepServer to be removed
   */
   public synchronized void shutdown() 
   {
      if (registeredWithDF() == true) 
      {
         // Attempt to deregister with DF
         try 
         {
            deregisterWithDF();
            DIAGNOSTICS.println("Deregistered with DF", this, DIAGNOSTICS.LEVEL_MAX );
         } catch (DFRegistrationException dfre) 
         {
            // A problem deregistering occured..
            String reason = dfre.getExceptionReason();
            DIAGNOSTICS.println(dfre + ":" + reason, this, DIAGNOSTICS.LEVEL_MAX ); 
         } // end catch
      } // if 
    
      if (registeredWithAMS() == true) 
      {
         // Attempt to deregister with AMS
         try 
         {
            deregisterWithAMS();
            DIAGNOSTICS.println("Deregistered with AMS", this, DIAGNOSTICS.LEVEL_MAX ); 

         } catch (AMSRegistrationException amsre) 
         {
            // A problem deregistering occured.
            String reason = amsre.getExceptionReason();
            DIAGNOSTICS.println(amsre + ":" + reason, this, DIAGNOSTICS.LEVEL_MAX );
         } // catch
      } // if 
     
      // Now call shutdown in the agent shell to release the core components
      super.shutdown();

   } // shutdown()
  

   /** retreive the replytoAID. Convenience method
   */
   public AgentID getReceiver(ACL acl)
   { 
      LinkedList ll = (LinkedList)acl.getReplyToAIDs();
      if (ll.size() > 0)
         return (AgentID)ll.get(0);
      else
         return null;
   }

   public int getConvs()
   {
      return getCM().getConvCount();
   }

   

   public void setLocalPostOffice( LocalPostOffice lpo)
   {
      _lpo = lpo; 
   }

   public LocalPostOffice getLocalPostOffice() 
   {
      return _lpo;
   }


   /** IdleTask. Listens for messages from the replicant group 
       or an outside agent. 
   */
   public class IdleTask extends Task 
   {


      public IdleTask() 
      {
      } // IdleTask()
    
    
      protected void startTask() 
      {
      } // StartTask()



      public void handleOther( Conversation conv) 
      {

         // An agent sent me a message so start 
         // a new conversation with ConvHandlerTask(). 
         newTask( new ConvHandlerTask(conv), conv);

      } // handleRequest()


      public void doneRepProxy_ConvHandlerTask(Task t)
      {
         // System.out.println( "RepProxy " + getAID().getName() + " Tasks: " + _tm.getTaskCount() + " Convs: " + getConvs() );
      }

   } // IdleTask()



   /** Handle a conversation 
   */
   public class ConvHandlerTask extends Task
   {

      /** the orginal message from the outside agent */
      private ACL _acl;
      private Conversation _inconv;
      private String _inconvid;

      /** the agent side */
      private Conversation _agentconv = null;
      private String       _agentconvid; 
      private AgentID      _agentid;

      /** the replicant group side */
      private Conversation _repconv = null;
      private String       _repconvid; 
      private AgentID      _repid;


      public ConvHandlerTask (Conversation conv)
      {
         _inconv = conv; 
         _inconvid = _inconv.getConversationID();
      } // InConvHandlerTask


      /** handle a new conversation 
      */
      protected void startTask()
      {
         // get the incoming message. 
         _acl = _inconv.getACL( _inconv.getLatestMessageIndex() );
         // see if the message has a ReplyTo in it. This is a kludge
         AgentID inc = getReceiver(_acl);

         // create a new message and a new conversation
         ACL forwd = getNewConversation( _acl.getProtocol() ); 
         String sendtoconvid = forwd.getConversationID();
         AgentID sendto;

         if ( inc == null) 
         { // this started from outside
            _agentconv   = _inconv;
            _agentconvid = _agentconv.getConversationID();
            _agentid     = _acl.getSenderAID();
            _repconvid   = sendtoconvid;
            _repid       = (AgentID) _repgroup.getActive(); 
            sendto       = _repid;
         }
         else 
         { // this started from the rep group
            _agentconvid = sendtoconvid; 
            _agentid     = inc; 
            _repconv     = _inconv;
            _repconvid   = _repconv.getConversationID();
            _repid       = _acl.getSenderAID();
            sendto       = _agentid; 
         }

         if (sendto == null) 
            return;

/*
         System.out.println("Message Proxy  : " + getAID().getName());
         if ( inc == null) 
         { // this started from outside
            System.out.println("Outside init:  " );
            System.out.println("From:          "  + _agentid.getName());
            System.out.println("To:            "  + _repid.getName());
         }
         else 
         {
            System.out.println("Inside init:  " );
            System.out.println("From:          "  + _repid.getName());
            System.out.println("To:            "  + _agentid.getName());
         }
         System.out.println("Performative:  "  + _acl.getPerformative());
         System.out.println("Ontology:      "  + _acl.getOntology());
         System.out.println("Content:       "  + _acl.getContentObject());
*/
      
         forwd.setReceiverAID( sendto );
         forwd.setSenderAID( getAID() );
         forwd.setPerformative( _acl.getPerformative() );
         forwd.setOntology( _acl.getOntology() );
         forwd.setContentObject( _acl.getContentObject() );
         forwd.setProtocol( _acl.getProtocol());
         forward( forwd );
     }


      public void handleOther(Conversation c) 
      {
         String thisconv = c.getConversationID();
         AgentID sendto; 
         String sendtoconvid; 
         
         ACL oth = c.getACL( c.getLatestMessageIndex());

         if ( thisconv.equals(_agentconvid) ) // if this is the agent side conversation...
         {
            sendto = _repid; 
            sendtoconvid = _repconvid; 
         } 
         else // this is the replicant group side of the conversation
         {
            sendto = _agentid;
            sendtoconvid = _agentconvid;
         }
      
         ACL forwd = new ACL(); 
         forwd.setConversationID( sendtoconvid );
         forwd.setReceiverAID( sendto );
         forwd.setSenderAID( getAID() );
         forwd.setPerformative( oth.getPerformative() );
         forwd.setOntology( oth.getOntology() );
         forwd.setContentObject( oth.getContentObject() );
         forwd.setProtocol( oth.getProtocol());
         forward( forwd );
   
         if ( c.getState() == -1 ) 
         {
            done();
         }

      } // handleOther()

   } // ConvHandlerTask

 

  /**
   * A main method to enable the agent to be run from the command line.
   * The agent takes two command line parameters - the platform profile location,
   * which is set up by the SetupFIPAOS.bat file to be in the system variable %PLATFORM%,
   * and the agent name.
   */
   public static void main( String args[] )
   {
		
   // Check for the correct number of arguments
   if( args.length < 2 )
   {
      System.out.println( "Usage: java " + RepProxy.class.getName() + " location-of-platform-profile agent-name" );
			
   }
   else
   {
      // args[ 0 ] = the platform profile location
      // args[ 1 ] = the agent name
      new RepProxy( args[ 0 ], args[ 1 ], "FIPA-OS", "test" , null);
   }
		
   }

} // RepProxy
