/* 
   Agent Server Testing - Agent Loader V1.0 

   Written by: 
      Alan Fedoruk                                           -- amf673@mail.usask.ca
      MADMUC Lab                                             -- http://bistrica.usask.ca/madmuc/index.htm
      Dept. of Computer Science, University of Saskatchewan. -- http://www.cs.usask.ca

   Copyright (c) 2001 Alan M. Fedoruk
*/
package testAS;

// Import the fipa-os classes
import fipaos.agent.*;
import fipaos.agent.conversation.*;
import fipaos.agent.task.*;
import fipaos.mts.*;
import fipaos.ont.fipa.*;
import fipaos.ont.fipa.fipaman.*;
import fipaos.parser.acl.ACLMessage;
import fipaos.platform.ams.AMSRegistrationException;
import fipaos.platform.df.DFRegistrationException;
import fipaos.util.DIAGNOSTICS;

// Need the repserver
import agentserver.*;

// java standard classes
import java.util.*;


/** Load Agents to test AgentServers
 */
public class LoadAgents extends FIPAOSAgent 
{
   /** A Vector of RepServer that this LoadAgents can send requests to */
   private Vector _agentservers;

   /** This Agent type */
   private static final String AGENT_TYPE = "loader-agent";

   /** Agent type to send requests to */
   private static final String RS_TYPE = ASCONSTANTS.AGENTSERVER; 
  
   /**  GUI for this TestAgent */
   private utils.AgentGUI _gui;
  

   /** Agent constructor - sets up an IdleTask as the default
       ListenerTask for incoming messages, and starts off registration
       with the AMS and DF */
   public LoadAgents( String platform, String name, String ownership ) 
   {
      super( platform, name, ownership );
    
      // Create a GUI for this Agent so we can see what is going on.
      try 
      {
         _gui = new utils.AgentGUI( getAID().getName() );
      } 
      catch ( Throwable t ) 
      { 
      }
    
      setListenerTask( new IdleTask() );
  
      // Now we can attempt to register with the local AMS (i.e. the platform)

      try 
      {
         registerWithAMS();
         DIAGNOSTICS.println( "Registered with AMS", this, DIAGNOSTICS.LEVEL_MAX );
         _gui.write( "Registered with AMS" );

      } 
      catch (AMSRegistrationException amsre)
      {
         // An exception has occured - this indicates that the AMS 

         DIAGNOSTICS.println(amsre, this, DIAGNOSTICS.LEVEL_MAX );
         String reason = amsre.getExceptionReason();
      
         // If the exception reason is  "already-registered", we can 
         // probably ignore the failure
         // (since this Agent was probably not de-registered the last time it ran!)

         if ( reason == null || !reason.equals(
               FIPAMANCONSTANTS.AGENT_ALREADY_REGISTERED)) 
         {
            shutdown();
            return;
         }
      } // catch 
    
      // Now we can register with the local DF
      try 
      {
         registerWithDF( AGENT_TYPE );
         DIAGNOSTICS.println("Registered with DF", this, 
               DIAGNOSTICS.LEVEL_MAX );
         _gui.write("Registered with DF" );
      } 
      catch (DFRegistrationException dfre) 
      {
         // An exception has occured - this indicates that 
         // the DF registration failed for some reason

         DIAGNOSTICS.println(dfre, this, DIAGNOSTICS.LEVEL_MAX );
      
         // We can easily find the exception reason from the exception....

         String reason = dfre.getExceptionReason();
      
         if ( reason == null || !reason.equals(
               FIPAMANCONSTANTS.AGENT_ALREADY_REGISTERED) ) 
         {

            shutdown();
            return;
         }
      
      } // end catch
    
   } // TestAgent()
  

   /** Invoked when this Agent is about to be shutdown - this overrides the
       shutdown() method in the FIPAOSAgent class to enable the TestAgentGUI
       associated with this TestAgent to be removed */
   public synchronized void shutdown() 
   {
      if (registeredWithDF() == true) 
      {
         // Attempt to deregister with DF
         try 
         {

            deregisterWithDF();
            DIAGNOSTICS.println("Deregistered with DF", this, 
                  DIAGNOSTICS.LEVEL_MAX );

         } catch (DFRegistrationException dfre) 
         {
            // A problem deregistering occured..... we can obtain 
            // the reason though!
            String reason = dfre.getExceptionReason();
            DIAGNOSTICS.println(dfre + ":" + reason, this, DIAGNOSTICS.LEVEL_MAX ); 
         } // end catch
      } // if 
    
      if (registeredWithAMS() == true) 
      {
         // Attempt to deregister with AMS
         try 
         {
            deregisterWithAMS();
            DIAGNOSTICS.println("Deregistered with AMS", this, 
                  DIAGNOSTICS.LEVEL_MAX ); 
         } 
         catch (AMSRegistrationException amsre) 
         {
            String reason = amsre.getExceptionReason();
            DIAGNOSTICS.println(amsre + ":" + reason, this, DIAGNOSTICS.LEVEL_MAX );
         } // catch
      } // if 
     
      // Dispose of the GUI

      if ( _gui != null ) 
      {
         _gui.dispose();
      }
    
      // Now call shutdown in the agent shell to release the core components
      super.shutdown();

   } // shutdown()
  

   /** This method is invoked when GUI should be shown - generally
       this is invoked by the AgentLoader when the Agent is in the List
       of running Agents - the default implementation in FIPAOSAgent does 
       nothing.  */
   public void activate() 
   {
      // Check if we've got a GUI, if so, show it
      if( _gui != null ) 
      {
         _gui.setVisible( true );
      }

   } // activate()
    


   /** Idletask, starts the loading process */
   public class IdleTask extends Task 
   {


      public IdleTask() 
      {
      } 
    
    
      protected void startTask() 
      {
         // Start the search for RS_TYPE
         findAgentServers();

      } // StartTask()

      public void findAgentServers()
      {
         // start the search for any agentservers
         DFAgentDescription df_desc = new DFAgentDescription();
      
         ServiceDescription sd = new ServiceDescription();
         sd.setServiceName(RS_TYPE);
         sd.setServiceType(RS_TYPE); 
         df_desc.setAgentServices( new ServiceDescription[] { sd } );
      
         // Start a search for other Agents
         searchDF( df_desc );
         _gui.write("Started DF Search for RepServers");
      }

    
      public void doneTestAgent_SendReqTask( Task t ) {
      } 
    

      /** Get the ids of the agentservers and put them in the 
          _agentservers vector.  */
      public void DFSearchResults( DFAgentDescription descs[] ) 
      {
         _agentservers = new Vector();
         // Check we've got some results - if not just return
         if( descs == null || descs.length <= 0 ) 
         {
            _gui.write("Could not find any agentservers. Try again in 30 s");
            // try again in 30 seconds. 
            newTask( new WaitTask(30000)); 
         }
         else 
         { 
            _gui.write( "Found " + (descs.length) + " repserver agents" );
        
            for ( int i = 0; i < descs.length; i++) 
            {
               AgentID aid = descs[i].getAgentID();
               _agentservers.addElement(aid);
            }
            // once we find some RepServers, send requests
            newTask( new SendReqTask());
         }
      
      } // DFSearchResults()
    
    
      /** If we Invoked when a WaitTask "completes" (i.e. after the period
          of time it has been asked to wait elapses).
      */
      public void doneWaitTask( Task t ) 
      {
         // 30 seconds is up, try again. 
         findAgentServers();  
      } // doneWaitTask()

   } // IdleTask()


   /** send requests to the AgentServers to test things */
   public class SendReqTask extends Task 
   {

      /* Counter for number of responses we're waiting for */
      private int _count;
      /* the number of CountAgent agents to create */ 
      private int _npas = 50;
      /* the number of replication servers in the system */
      private int _nrs;


      public SendReqTask(  ) 
      {
      }


      protected void startTask() 
      {
         // send a request to create a matchmaker
         _nrs = _agentservers.size();
         AgentID agentserver = (AgentID)_agentservers.elementAt( 0); 
         ACL acl = getNewConversation( FIPACONSTANTS.FIPA_REQUEST );
         acl.setPerformative( FIPACONSTANTS.REQUEST );
         acl.setSenderAID( _owner.getAID() );
         acl.setReceiverAID( agentserver );
         acl.setOntology( ASCONSTANTS.AGENTSERVER_ONT);
         acl.setContentObject( ASCONSTANTS.CREATE + " agentserver.CountAgent c " + FIPACONSTANTS.FIPA_OS );
         forward( acl );
         _gui.write("Sent request: " + acl.getContentObject() );

      } // startTask()


      public void handleInform(Conversation conv) 
      {
         // when the previous request is finished, wait 10 seconds. 
         _gui.write( "Done Request!");
         newTask( new WaitTask(10000));
      }


      public void handleAgree(Conversation conv)
      {
         _gui.write("Agreed!");
      }


      public void handleRefuse(Conversation conv)
      {
         _gui.write("Refused! --- go on to next ");
         newTask( new WaitTask(10000));
      }


      public void handleFailure(Conversation conv)
      {
         _gui.write("Failed! --- go on to next");
         newTask( new WaitTask(10000));
      }


      public void doneWaitTask(Task t)
      {
         // 10 seconds is up. If we have done enough PersAgent stop. Otherwise send another 
         // on PersAgent RepGroup per RepServer
         if ( _count < _npas) 
         { 
            AgentID repserver = (AgentID)_agentservers.elementAt( _count % _nrs); 
            ACL acl = getNewConversation( FIPACONSTANTS.FIPA_REQUEST );

            acl.setPerformative( FIPACONSTANTS.REQUEST );
            acl.setSenderAID( _owner.getAID() );
            acl.setReceiverAID( repserver );
            acl.setOntology( ASCONSTANTS.AGENTSERVER_ONT);
            acl.setContentObject( ASCONSTANTS.CREATE + " agentserver.CountAgent c" + _count + " " + FIPACONSTANTS.FIPA_OS );
            forward( acl );
            _gui.write("Sent request: " + acl.getContentObject() );
            _count++; 
         }
         else 
         {
            done();
         }
      }

   
   } // ReqAllTask
   

   /**
   * Test-harness
   */
   public static void main( String args[] ) 
   {
      if ( args == null || args.length < 2 ) 
      {
         System.out.println( "Usage: java testRS.TestAgent <platform.profile> <name>" );
         return;
      }
      new LoadAgents( args[0], args[1], "FIPA-OS" );

   } // main()

} // TestAgent
