/* 
   Test Replication Server - TestAgent V1.0, November 6, 2001 

   Written by: 
      Alan Fedoruk                                           -- amf673@mail.usask.ca
      MADMUC Lab                                             -- http://bistrica.usask.ca/madmuc/index.htm
      Dept. of Computer Science, University of Saskatchewan. -- http://www.cs.usask.ca

   Copyright (c) 2001 Alan M. Fedoruk
*/


  
package testRS;

// Import the fipa-os classes
import fipaos.agent.*;
import fipaos.agent.conversation.*;
import fipaos.agent.task.*;
import fipaos.mts.*;
import fipaos.ont.fipa.*;
import fipaos.ont.fipa.fipaman.*;
import fipaos.parser.acl.ACLMessage;

// Import registration exception classes
import fipaos.platform.ams.AMSRegistrationException;
import fipaos.platform.df.DFRegistrationException;

import fipaos.util.DIAGNOSTICS;

import repserver.*;

import java.util.*;

/** TestAgent exercises the replication server.
    It sends requests to the replication server to create two replicant groups, 
    then it sends requests to the counter replicant group. 
 */
public class TestAgent extends FIPAOSAgent 
{
   /** A RepServer that this TestAgent can send requests to */
   private AgentID _rs_agent;

   /** This Agent type */
   private static final String AGENT_TYPE = "test-agent";

   /** Agent type to send requests to */
   private static final String RS_TYPE = RSCONSTANTS.REPSERVER; 
  
   /**  GUI for this TestAgent */
   private AgentGUI _gui;
  

   /**
    * Agent constructor - sets up an IdleTask as the default
    * ListenerTask for incoming messages, and starts off registration
    * with the AMS and DF
    */
   public TestAgent( String platform, String name, String ownership ) 
   {
      super( platform, name, ownership );
 
      // Create a GUI for this Agent so we can see what is going on.
      try 
      {
         _gui = new AgentGUI( getAID().getName() );
      } 
      catch ( Throwable t ) 
      { 
      }
    
      // Instantiate a new IdleTask to receieve all incoming 
      // communications which are not associated with any-other Task

      setListenerTask( new IdleTask() );
  
      try 
      {
         registerWithAMS();
         DIAGNOSTICS.println( "Registered with AMS", this, DIAGNOSTICS.LEVEL_MAX );
         _gui.write( "Registered with AMS" );

      } 
      catch (AMSRegistrationException amsre)
      {
         DIAGNOSTICS.println(amsre, this, DIAGNOSTICS.LEVEL_MAX );
         String reason = amsre.getExceptionReason();
         if ( reason == null || !reason.equals(
               FIPAMANCONSTANTS.AGENT_ALREADY_REGISTERED)) 
         {
            // Shutdown nicely :)
            shutdown();
            return;
         }
      } // catch 
    
      try 
      {
         registerWithDF( AGENT_TYPE );
         DIAGNOSTICS.println("Registered with DF", this, 
               DIAGNOSTICS.LEVEL_MAX );
         _gui.write("Registered with DF" );
      } 
      catch (DFRegistrationException dfre) 
      {
         DIAGNOSTICS.println(dfre, this, DIAGNOSTICS.LEVEL_MAX );
         String reason = dfre.getExceptionReason();
         if ( reason == null || !reason.equals(
               FIPAMANCONSTANTS.AGENT_ALREADY_REGISTERED) ) 
         {

            shutdown();
            return;
         }
      
      } // end catch
    
   } // TestAgent()
  

   /**
    * Invoked when this Agent is about to be shutdown - this overrides the
    * shutdown() method in the FIPAOSAgent class to enable the TestAgentGUI
    * associated with this TestAgent to be removed
    */
   public synchronized void shutdown() 
   {
      if (registeredWithDF() == true) 
      {
         // Attempt to deregister with DF
         try 
         {

            deregisterWithDF();
            DIAGNOSTICS.println("Deregistered with DF", this, 
                  DIAGNOSTICS.LEVEL_MAX );

         } catch (DFRegistrationException dfre) 
         {
            // A problem deregistering occured..... we can obtain 
            // the reason though!
            String reason = dfre.getExceptionReason();

            DIAGNOSTICS.println(dfre + ":" + reason, this, 
                  DIAGNOSTICS.LEVEL_MAX ); 
         } // end catch
      } // if 
    
      // Check if we're registered with the AMS
      if (registeredWithAMS() == true) 
      {
         // Attempt to deregister with AMS
         try 
         {
            deregisterWithAMS();
            DIAGNOSTICS.println("Deregistered with AMS", this, 
                  DIAGNOSTICS.LEVEL_MAX ); 
         } 
         catch (AMSRegistrationException amsre) 
         {
            // A problem deregistering occured..... we can obtain the 
            // reason though!
            String reason = amsre.getExceptionReason();
     
            DIAGNOSTICS.println(amsre + ":" + reason, this, DIAGNOSTICS.LEVEL_MAX );
         } // catch
      } // if 
     
      // Dispose of the GUI

      if ( _gui != null ) 
      {
         _gui.dispose();
      }
    
      // Now call shutdown in the agent shell to release the core components
      super.shutdown();

   } // shutdown()
  

   /**
    * This method is invoked when GUI should be shown - generally
    * this is invoked by the AgentLoader when the Agent is in the List
    * of running Agents - the default implementation in FIPAOSAgent does 
    * nothing.
    */

   public void activate() 
   {
      // Check if we've got a GUI, if so, show it
      if( _gui != null ) 
      {
         _gui.setVisible( true );
      }

   } // activate()
    


   public class IdleTask extends Task 
   {


      public IdleTask() 
      {
      } 
    
    
      protected void startTask() 
      {
         // Start the search for RS_TYPE
   
         // Create the DFAgentDescription describing the attributes we wish
         // to constrain in our search for DataAgents
         DFAgentDescription df_desc = new DFAgentDescription();
      
         // Create a ServiceDescription of the replication server
         ServiceDescription sd = new ServiceDescription();
         sd.setServiceName(RS_TYPE);
         sd.setServiceType(RS_TYPE); 
         df_desc.setAgentServices( new ServiceDescription[] { sd } );
      
         // Start a search for other Agents
         searchDF( df_desc );
         DIAGNOSTICS.println( "TestAgent: Started DF search", DIAGNOSTICS.LEVEL_4);
         _gui.write("Started DF Search");
   
         newTask( new WaitTask( 15000 ) );
         DIAGNOSTICS.println("TestAgent: Started WaitTask. When it is done start sending requests.", 
               DIAGNOSTICS.LEVEL_4 );
      } // StartTask()
    
    
      public void doneTestAgent_SendReqTask( Task t ) {
      } 
    
      /** This method is invoked when a DFSearchTask has completed
          succesfully.
       */
      public void DFSearchResults( DFAgentDescription descs[] ) 
      {
         // Check we've got some results - if not just return
         if( descs == null || descs.length <= 0 ) 
         {
            return;
         }
         DIAGNOSTICS.println( "TestAgent: Found " + (descs.length) + " RepServer agents", DIAGNOSTICS.LEVEL_3 );
         _gui.write( "Found " + (descs.length) + " data agents" );
        
         _rs_agent = descs[0].getAgentID();
      
      } // DFSearchResults()
    
    
      /** Invoked when a WaitTask "completes" (i.e. after the period
          of time it has been asked to wait elapses).
      */
      public void doneWaitTask( Task t ) 
      {
         // Indicate what is going on
         DIAGNOSTICS.println("TestAgent: doneWaitTask() has been called.\n"+
               "Send request to the replication server.", DIAGNOSTICS.LEVEL_3 );
  
         newTask( new SendReqTask( _rs_agent) );
         // start sending requests to the CountAgent
         newTask( new GetNTask());

      } // doneWaitTask()


      public void doneTestAgent_GetNTask( Task t) 
      {
      }

     
   } // IdleTask()


   /** Sends 10 requests, 10 seconds apart to the counter agent. 
       and displays the results.
   */
   public class GetNTask extends Task
   {

      private int _count = 10;


      public GetNTask()
      {
      }
   

      protected void startTask()
      {
         newTask( new WaitTask(10000)); 
      }


      public void doneWaitTask( Task t) 
      {
         // waiting is over, send the request. 
         ACL acl = new ACL(); 
         acl.setPerformative(FIPACONSTANTS.REQUEST);
         acl.setSenderAID(_owner.getAID());

         // could put in a lookup here. 
         AgentID a = new AgentID(); a.setName("counter@amf673");
         acl.setReceiverAID( a);
         acl.setOntology("none"); 
         forward(acl);

         // if we have done all we need to, stop
         _count--;
         if (_count > 0) 
            newTask( new WaitTask(10000));
         else 
            done();
      }
 

      public void handleAgree(Conversation conv) 
      {
         // Display the fact that we got an agree
         ACL acl = conv.getACL(conv.getLatestMessageIndex());
         _gui.write("From Agent " + acl.getSenderAID().getName() + " got Agree");
      }


      public void handleInform(Conversation conv)
      {
         // Display the content of the INFORM
         ACL acl = conv.getACL(conv.getLatestMessageIndex());
         _gui.write("From Agent " + acl.getSenderAID().getName() + " " + acl.getContentObject());
      }

   }


   /** Send a replication request to the replication server. 
   */
   public class SendReqTask extends Task 
   {

      /** The Agents to send request to
      */
      private AgentID _rs_agent;

      /** Counter for number of responses we're waiting for */
      private int _count = 2;

      /** Constructs a ReqAllTask, which has the purpose of sending a request to each of the
          Agent specified in the given List.
      */
      public SendReqTask( AgentID agent ) 
      {
         _rs_agent = agent;
      }


      protected void startTask() 
      {
         AgentID a = new AgentID(); 
         a.setName("repserver0@amf673");

         ACL acl1 = getNewConversation( FIPACONSTANTS.FIPA_REQUEST );
         acl1.setPerformative( FIPACONSTANTS.REQUEST );
         acl1.setSenderAID( _owner.getAID() );
         acl1.setReceiverAID( _rs_agent );
         acl1.setOntology( RSCONSTANTS.REPSERVER_ONT);
         acl1.setContentObject( RSCONSTANTS.REPLICATE + " counter testRS.CountAgent counter 3" );
         forward( acl1 );
         _gui.write("Sent request to replicate CountAgent as counter, to " +_rs_agent);
/*

         ACL acl2 = new ACL(); 
         acl2.setPerformative( FIPACONSTANTS.REQUEST );
         acl2.setSenderAID( _owner.getAID() );
         acl2.setReceiverAID( _rs_agent );
         acl2.setOntology( RSCONSTANTS.REPSERVER_ONT);
         acl2.setContentObject( RSCONSTANTS.REPLICATE + " enumerator testRS.CountAgent counter 3" );
         forward( acl2 );
         _gui.write("Sent request to replicate CountAgent as enumerator, to " +_rs_agent);
*/

      } // startTask()


      public void handleAgree(Conversation conv) 
      {
         decrementCount();
      }

      public synchronized void decrementCount() 
      {
         _count--;
         if ( _count == 0 ) {
            // If _counter == 0, we're done!
            done();
         }
      } // decrementCount()
   
   } // ReqAllTask
   

   /**
   * Test-harness
   */
   public static void main( String args[] ) 
   {
      if ( args == null || args.length < 2 ) 
      {
         System.out.println( "Usage: java testRS.TestAgent <platform.profile> <name>" );
         return;
      }
      new TestAgent( args[0], args[1], "FIPA-OS" );

   } // main()

} // TestAgent
