/* 
   Replication Server - RepAgent V1.1, November 20, 2001 

   Written by: 
      Alan Fedoruk                                           -- amf673@mail.usask.ca
      MADMUC Lab                                             -- http://bistrica.usask.ca/madmuc/index.htm
      Dept. of Computer Science, University of Saskatchewan. -- http://www.cs.usask.ca

   Copyright (c) 2001 Alan M. Fedoruk
*/

package repserver;

// Import the fipa-os classes
import fipaos.agent.*;
import fipaos.agent.task.*;
import fipaos.ont.fipa.fipaman.*;
import fipaos.ont.fipa.*;
import fipaos.agent.conversation.*;
import fipaos.util.DIAGNOSTICS;

// Import java classes
import java.util.StringTokenizer;
import java.util.*;

// Import local classes
import localpostoffice.*; 

/** RepAgent is an abstract class representing an agent shell suitable for 
    use with a replication server. It is an extension of the standard 
    FIPAOSAgent shell. It adds variables to store the id of the 
    replication server associated with this agent, the name of the 
    replicant group this agent belongs to, the agentID of the message 
    proxy for the replicant group and the current role of the agent. 
    <P> 
    Responses to various repserver-management tasks are included. 
    The person using this is responsible for including an appropriate 
    set of code in the agents IDLECLASS which will call the message 
    handlers.  
    <P>
    A setAState and getAState method stub is included in this abstract 
    class. The user must override these as appropriate. 

*/
public abstract class RepAgent extends FIPAOSAgent implements LocalMsg
{

   /** the AID of the repserver associated with this agent */
   private AgentID _repserver = null;   

   /** the name of the replicant group that I will belong to */
   private String _group_name = null;

   /** the AgentID of my mail proxy. */
   private AgentID _group_proxy = null; 

   /** the current role of this agent. Default to REPLICANT */
   private String _role = RSCONSTANTS.REPLICANT; 
  
   /** the localPostOffice if desired */
   private LocalPostOffice _lpo = null; 


   /** Constructor. Call the FIPAOSAgent constructor but 
       also set the role, group name and replication server.  
   */
   public RepAgent( String platform_profile_location, String agent_name,
	String ownership, AgentID rs, String role, String group_name) 
   {
      super(platform_profile_location, agent_name, ownership, new RepAgentComponentFactory() );
      _repserver = rs; 
      _group_name = group_name;
      setRole(role); 

      // if the agent knows the _repserver, then send a registration 
      // message via a RegisterTask. 

      if ( _repserver != null ) 
      { 
         _tm.newTask( new RegisterTask() );
      }

   } // RepAgent()


   public void setLocalPostOffice( LocalPostOffice lpo)
   {
      _lpo = lpo; 
   }


   public LocalPostOffice getLocalPostOffice() 
   {
      return _lpo;
   }


   /** Returns true if this agent is currently in ACTIVE_AGENT role 
   */
   public boolean isActiveAgent() 
   {
      if ( getRole().equals(RSCONSTANTS.ACTIVE_AGENT)) 
         return true; 
      
      return false; 

   } // isReplicant()


   /** Returns true if this agent is currently in REPLICANT role 
   */
   public boolean isReplicant() 
   {
      if ( getRole().equals(RSCONSTANTS.REPLICANT)) 
         return true; 
      
      return false; 

   } // isReplicant()

   
   /** Returns the current role of this REPLICANT
   */
   public String getRole() 
   {
      return _role; 
   } 


   /** Sets the role of this agent. If the switch is from REPLICANT to 
       ACTIVE_AGENT, call the goActive() method
       @param role String representing the role, see RSCONSTANTS. 
   */
   public boolean setRole(String role) 
   {

      if ( role.equals(RSCONSTANTS.REPLICANT) || 
            role.equals(RSCONSTANTS.ACTIVE_AGENT)) 
      {
         if ( _role.equals(RSCONSTANTS.REPLICANT) && 
               role.equals(RSCONSTANTS.ACTIVE_AGENT) ) 
         {
            // switching to active role, call goActive method
            _role = role; 
            goActive();
         }
         _role = role;
         return true; 
      } 
      else 
      {
         return false; 
      }

   } // setRole()


   /** Returns the AgentID of this agents RepServer 
   */
   public AgentID getRepServer() 
   {
      return _repserver; 
   } //getRepServer()


   /* Set the RepServer to be AgentID
      @param aid AgentID of new replication server. 
   */
   public void setRepServer(AgentID aid) 
   {
      _repserver = aid;

      if ( _repserver != null ) 
      { 
         _tm.newTask( new RegisterTask() );
      }

   } // setRepServer()

   public void doneRepAgent_RegisterTask( Task t)
   {
   }

   /** Return the group name of the replicant group this agent is registered with
   */
   public String getGroupName() 
   {
      return _group_name; 
   }


   /** Return the AgentID of the message proxy this agent is registered with
   */
   public AgentID getProxyAID()
   {
      return _group_proxy;
   }


   /** Code to execute when this agent is set to ACTIVE_AGENT. 
       This method should be overridden when implementing an agent. 
   */
   public void goActive()
   {
   }


   /** returns a byte stream representing the current state of this 
       agent. This should be overridden by the implementing agent.
   */
   public byte[] getAState() 
   {
      return ( new byte[1]); 
   } 


   /** sets the state for this agent. 
       @param state a byte stream representing the state of this agent. 
       returns true if successful.
   */
   public boolean setAState(byte[] state) 
   {
      return true; 
   } //setAState


   /** retreive the replytoAID. Convenience method
   */
   public AgentID getReceiver(ACL acl)
   { 
      LinkedList ll = (LinkedList)acl.getReplyToAIDs();
      if (ll.size() > 0)
         return (AgentID)ll.get(0);
      else
         return null;
   }


   /** Send a reply message in response to the given ACL message
       @param acl The ACL object to reply to
       @param reason Any additional message
   */
   public boolean sendReply( ACL orig_acl, String performative, String content )
   {
      try
      {
         // Duplicate ACL
         ACL acl = (ACL) orig_acl.clone();
            
         // Setup message
         acl.setPerformative( performative );
         acl.setReceiverAID( acl.getSenderAID() );
         acl.setSenderAID( getAID() );
         acl.setInReplyTo( acl.getReplyWith() );
         acl.setReplyWith( null );
            
         // Set content
         acl.setContentObject( content );
            
         // Attempt to send the reply
         forward( acl );
            
         // We're done
         return true;
      }
      catch( Throwable t )
      {
         DIAGNOSTICS.println( "Error occured when sending reply", t, this,
         DIAGNOSTICS.LEVEL_MAX );
         return false;
      }
   } // sendReply()

   
   public boolean sendReply( ACL orig_acl, String performative)
   {
      return sendReply(orig_acl, performative, ""); 
   }


   /** Get the number of current conversations */
   public int getConvs()
   {
      return getCM().getConvCount();
   }


   /** remove the given conversation */
   public void removeConv(Conversation conv)
   {
      try 
      {
         getCM().removeConversation(conv.getConversationID());
      }
      catch (Exception e)
      {
         System.out.println(e);
      }
   }





   /** RegisterTask, registers an agent with a RepGroup 
   */
   public class RegisterTask extends Task 
   {

      public RegisterTask() 
      {
      } // RegisterTask()


      protected void startTask() 
      {
         // Send a register message to RepServer to say.. I am here! 
         ACL reg = new ACL( );
	 reg.setSenderAID( getAID() );
	 reg.setReceiverAID( _repserver );
	 reg.setPerformative( FIPACONSTANTS.REQUEST );
         reg.setOntology( RSCONSTANTS.REPSERVER_ONT);
         reg.setContentObject( RSCONSTANTS.REGISTER + " " + _group_name);
         reg.setProtocol(FIPACONSTANTS.FIPA_REQUEST);
	 forward( reg );

      } // startTask()


      public void handleAgree(Conversation conv) 
      {
         // Just accept the Agree.
      }


      public void handleInform(Conversation conv)
      {
         // set the proxy agentID
         ACL inform = conv.getACL( conv.getLatestMessageIndex()); 
         _group_proxy = (AgentID) inform.getContentObject();
         done();
      }

   } // RegisterTask


   /** Handles other RepServerManagementTask functions. 
       The extending agent should have code in its IdleTask like: 

       public void handleRequest( Conversation conv ) {
         // Get the type of request
         ACL acl = conv.getACL( conv.getLatestMessageIndex() );
         // Check ontology
         if(acl.getOntology().equalsIgnoreCase(RSCONSTANTS.REPSERVER_ONT)) {
            newTask ( new RepServerManagementTask(conv, acl), conv); 
	 } else {
            sendNotUnderstood(acl);
	    return;
         }
      } // handleRequest()

      which for all incoming requests with ontology of REPSERVER_ONT 
      start one of these tasks. 
   */
   public class RepServerManagementTask  extends Task 
   {

      private ACL _acl; 


      /** Constructor. Set the conv and acl. 
      */
      public RepServerManagementTask(ACL acl) 
      {
         _acl = acl; 
      } 
    

      protected void startTask() 
      {
         // determine if this is a request... 

         if ((_acl.getPerformative()).equals(FIPACONSTANTS.REQUEST)) 
         {
            StringTokenizer st = new StringTokenizer( 
                  (String)_acl.getContentObject());
            String function = st.nextToken(); 

            if ( function.equals( RSCONSTANTS.PING )) 
            {
               // send an agree to show that we got the message 
               sendReply( _acl, FIPACONSTANTS.AGREE);
               sendReply( _acl, FIPACONSTANTS.INFORM);
               done();
            } 
            else if (function.equals( RSCONSTANTS.LOADSTATE )) 
            {
               // this is a LOADSTATE, 
               // send an agree to show that we got the message 
               // an INFORM will come shortly
               sendReply( _acl, FIPACONSTANTS.AGREE);
            } 
            else if (function.equals( RSCONSTANTS.ACTIVATE )) 
            {
               // request to ACTIVATE this agent
               setRole( RSCONSTANTS.ACTIVE_AGENT);
               sendReply( _acl, FIPACONSTANTS.AGREE);
               sendReply( _acl, FIPACONSTANTS.INFORM);
               done();
            } 
            else 
            {   
               // otherwise send a not understood. 
               sendNotUnderstood(_acl);
               done();
            }
         }

      } // startTask()


      public void handleAgree(Conversation conv) 
      {
      }


      public void handleInform(Conversation conv) 
      {
         if ( !getRole().equals(RSCONSTANTS.ACTIVE_AGENT)) 
         {
            // the INFORM will hold the state information in its Content field. 
            ACL acl = conv.getACL( conv.getLatestMessageIndex() );
            byte[] state = (byte[])acl.getContentObject(); 
            setAState(state); 
         } 
         removeConv(conv);
         done();
      }

   } // RepServerManagementTask


   /** StateTask sends a request to the repserver, 
       then when an AGREE comes back send an INFORM with the 
       agent state in the content field. 
   */
   public class StateTask extends Task 
   {

      private byte[] _state; 

      public StateTask() 
      {
         _state = getAState(); 
      } 
   

      protected void startTask() 
      {
         // DIAGNOSTICS.println("StateTask: Sending REQUEST STATE", DIAGNOSTICS.LEVEL_4);
         // Send a STATE message to RepServer to say.. I am going to send 
         // my state. 
         ACL resp = new ACL( );
	 resp.setSenderAID( getAID() );
	 resp.setReceiverAID( _repserver );
	 resp.setPerformative( FIPACONSTANTS.REQUEST );
         resp.setOntology( RSCONSTANTS.REPSERVER_ONT);
         resp.setContentObject( RSCONSTANTS.STATE + " " + _group_name);
	 forward( resp );
          
      } // startTask()


      public void handleAgree(Conversation conv) 
      {
         // DIAGNOSTICS.println("StateTask: Sending INFORM ", DIAGNOSTICS.LEVEL_4);
         ACL acl = conv.getACL( conv.getLatestMessageIndex() );
         ACL resp =  (ACL)acl.clone(); 
	 resp.setSenderAID( getAID() );
	 resp.setReceiverAID( _repserver );
	 resp.setPerformative( FIPACONSTANTS.INFORM );
         resp.setOntology( RSCONSTANTS.REPSERVER_ONT);
         resp.setContentObject( _state);
	 forward( resp );
         removeConv( conv );
         done();
         
      } // handleAgree

   } // StateTask
    
} // RepAgent
