/* Replication Server RepServer V1.1, November 20, 2001  

   Written by: 
      Alan Fedoruk                                           -- amf673@mail.usask.ca
      MADMUC Lab                                             -- http://bistrica.usask.ca/madmuc/index.htm
      Dept. of Computer Science, University of Saskatchewan. -- http://www.cs.usask.ca

   Copyright (c) 2001 Alan M. Fedoruk

   Modifications: 

   Nov 20, 2001 - Added getConvs and removeConv methods. 
                - Added sendReply method. 
                - Conversations were not getting cleaned up. 
   Dec 06, 2001  - Fixed support for COLD-STANDBY
*/

package repserver;

// Import the fipa-os classes
import fipaos.agent.*;
import fipaos.agent.conversation.*;
import fipaos.agent.task.*;
import fipaos.mts.*;
import fipaos.ont.fipa.*;
import fipaos.ont.fipa.fipaman.*;
import fipaos.parser.acl.ACLMessage;
import fipaos.util.DynamicClass;

// Import registration exception classes
import fipaos.platform.ams.AMSRegistrationException;
import fipaos.platform.df.DFRegistrationException;
import fipaos.util.DIAGNOSTICS;

import java.util.*;

import localpostoffice.*;


/** RepServer implements a proxy for replicant groups. 
    <P>
    The RepServer extends the FIPAOSAgent shell. 
    <P>
    RepServer manages a set of replicant groups (RepGroup). Each group consisting
    of a list of the agentids making up the group. The RepServer manages updating 
    The repserver performs three tasks: 
    <UL>
       <LI>respond to incoming requests. 
       <LI>ping the agents in the replicant groups to ensure they are still alive. 
       <LI>update the states of the agents in each replicant group. 
    </UL>
    All requests are sent using the repserver-management ontology. Any other 
    ontology is ignored.  The possible requests are: 
    <UL>
       <LI>replicate   -- create a new replicant group. 
       <LI>register    -- register the sending agent with an exising replicant group. 
                     the sending agent can exist on any platform, within this RepServer 
                     or in another RepServer.
       <LI>state       -- get the state from an active agent. 
       <LI>create      -- create an agent and add it to an existing replicant group. 
   </UL>
    When a replicate request is recieved, or a create, agents are created in this process space. 
    Agents seeking to register can come from any process space. 
        
*/
public class RepServer extends FIPAOSAgent implements LocalMsg
{

   /** This Agent type */
   private static final String AGENT_TYPE = RSCONSTANTS.REPSERVER;

   /** List of agents that this repserver started */
   private Hashtable _started_agents = new Hashtable(); 

   /** Hashtable of all the RepGroups that this server has */ 
   private Hashtable _repgroups;

   /**  GUI for this ReqAgent */
   private RepServerGUI _gui;

   /** Location of platform profile for use when creating agents */
   private String _platform_profile;

   private LocalPostOffice _localpo = null; 

   /** a reference to this repserver */
   private RepServer _repserver = this;

   /** constructor - sets up an IdleTask as the default
       ListenerTask for incoming messages, and starts off registration
       with the AMS and DF
   */
   public RepServer (String platform, String name, String ownership, String gui ) 
   {
   
      super( platform, name, ownership, new RepAgentComponentFactory() );
    
      _platform_profile = platform;
      _repgroups = new Hashtable();

      _localpo = null; // new LocalPostOffice(_started_agents);
      _started_agents.put( getAID().getName(), this);

      // Create a GUI for this Agent so we can see what is going on.
      if ( gui.equals("gui"))
      {
         try 
         {
            _gui = new RepServerGUI( getAID().getName(), this );
         } 
         catch ( Throwable t ) 
         { 
            System.out.println("Cannot Create GUI!");
         }
      }
    
      // Instantiate a new IdleTask to receieve all incoming 
      // communications which are not associated with any other Task

      setListenerTask( new IdleTask() );
  
      // Register with the local AMS (i.e. the platform)
      try 
      {

         registerWithAMS();
         writeMsg("Registered with AMS" );

      } 
      catch ( AMSRegistrationException amsre ) 
      {
         String reason = amsre.getExceptionReason();
      
         if ( reason == null || !reason.equals( FIPAMANCONSTANTS.AGENT_ALREADY_REGISTERED)) 
         {
            shutdown();
            return;
         }
         writeMsg("Already Registered with AMS" );
      
      } // end catch 
    
      // Register with the local DF
      try 
      {
         registerWithDF( AGENT_TYPE );
         writeMsg("Registered with DF" );

      } 
      catch (DFRegistrationException dfre) 
      {
      
         String reason = dfre.getExceptionReason();
      
         if ( reason == null || !reason.equals( FIPAMANCONSTANTS.AGENT_ALREADY_REGISTERED) ) 
         {
            shutdown();
            return;
         }
         writeMsg("Already Registered with DF" );
      } // end catch

     // start a PingUpdateTask which manages agent pings and agent state updates. 
     _tm.newTask( new PingUpdateTask() );
    
   } // RepServer()
  

   /** Invoked when this Agent is about to be shutdown - this overrides the
       shutdown() method in the FIPAOSAgent class to enable the RepServerGUI
       associated with this RepServer to be removed
   */
   public synchronized void shutdown() 
   {
      if (registeredWithDF() == true) 
      {
         // Attempt to deregister with DF
         try 
         {
            deregisterWithDF();
         } 
         catch (DFRegistrationException dfre) 
         {
            // A problem deregistering occured..... 
            String reason = dfre.getExceptionReason();

         } // end catch
      } // if 
    
      if (registeredWithAMS() == true) 
      {
         // Attempt to deregister with AMS
         try 
         {
            deregisterWithAMS();

         } 
         catch (AMSRegistrationException amsre) 
         {
            // A problem deregistering occured.....
            String reason = amsre.getExceptionReason();
     
         } // catch
      } // if 
     
      // Dispose of the GUI
      if( _gui != null ) 
      {
         _gui.dispose();
      }

      // send a shutdown to any message proxies started in this process. 

      for (Enumeration e = _repgroups.keys(); e.hasMoreElements() ;) 
      {
         RepGroup rg = (RepGroup)_repgroups.get( (String)e.nextElement());
         rg.getProxy().shutdown();
      }
      // send a shutdown to any agents started in this process. 

      for (Enumeration e = _started_agents.keys(); e.hasMoreElements() ;) 
      {
         RepAgent ra = (RepAgent)_started_agents.get( (String)e.nextElement());
         ra.shutdown();
      }

      // Now call shutdown in the agent shell to release the core components
      super.shutdown();

   } // shutdown()
  

   /** This method is invoked when GUI should be shown - generally
       this is invoked by the AgentLoader when the Agent is in the List
       of running Agents - the default implementation in FIPAOSAgent does 
       nothing.
   */
   public void activate() 
   {
      // Check if we've got a GUI, if so, show it
      if( _gui != null ) 
      {
         _gui.setVisible( true );
      }

   } // activate()
    

   /** write a message, either to the GUI or to system.out
   */
   public void writeMsg( String msg)
   {
      if( _gui != null ) 
         _gui.write( msg);
      else 
         System.out.println( msg);
   }


   public void setLocalPostOffice( LocalPostOffice lpo)
   {
      _localpo = lpo; 
   }

   public LocalPostOffice getLocalPostOffice() 
   {
      return _localpo;
   }


   public int getConvs()
   {
      return getCM().getConvCount();
   }


   public void removeConv(Conversation conv)
   {
      try 
      {
         getCM().removeConversation(conv.getConversationID());
      }
      catch (Exception e)
      {
         System.out.println(e);
      }
   }



   public boolean sendReply( ACL orig_acl, String performative, Object content )
   {
      try
      {
         // Duplicate ACL
         ACL acl = (ACL) orig_acl.clone();
            
         // Setup message
         acl.setPerformative( performative );
         acl.setReceiverAID( acl.getSenderAID() );
         acl.setSenderAID( getAID() );
         acl.setInReplyTo( acl.getReplyWith() );
         acl.setReplyWith( null );
            
         // Set content
         acl.setContentObject( content );
            
         // Attempt to send the reply
         forward( acl );
            
         // We're done
         return true;
      }
      catch( Throwable t )
      {
         DIAGNOSTICS.println( "Error occured when sending reply", t, this,
         DIAGNOSTICS.LEVEL_MAX );
         return false;
      }
   } // sendReply()

   
   public boolean sendReply( ACL orig_acl, String performative)
   {
      return sendReply(orig_acl, performative, (Object) ""); 
   }
	

   /** manages the PingRepGroupsTask tasks. 
   */
   public class PingUpdateTask extends Task 
   {
      

      public PingUpdateTask() 
      {
      }


      protected void startTask() 
      {
         newTask( new WaitTask(RSCONSTANTS.SECONDS_5)); 

      }


      public void doneWaitTask( Task t) 
      {
         // if there are any repgroups, start a PingRepGroupsTask
         // if there are no repgroups, wait a bit and check again. 
         // writeMsg("RepServer " + getAID().getName() + " Tasks: " + _tm.getTaskCount() + " Convs: " + getConvs());
         if ( _repgroups.size() > 0 ) 
            newTask( new PingRepGroupsTask());
         else 
            newTask( new WaitTask(RSCONSTANTS.SECONDS_10)); 

      } // doneWaitTask()


      public void doneRepServer_PingRepGroupsTask( Task t) 
      {
         newTask( new WaitTask(RSCONSTANTS.SECONDS_10)); 

      } // doneRepServer_PingRepGroupsTask


      public void timeoutRepServer_PingRepGroupsTask( Task t) 
      {
         newTask( new WaitTask(RSCONSTANTS.SECONDS_10)); 

      } // timeoutRepServer_PingRepGroupsTask
   
   } // PingUpdateTask




   /** IdleTask. Handles all processing of messages not associated with any 
       other conversation. 
   */
   public class IdleTask extends Task 
   {


      public IdleTask() 
      {
      } // IdleTask()
    
    
      protected void startTask() 
      {
      } // StartTask()


      public void handleRequest( Conversation conv) 
      {
         // Get the type of request by checking ontology 
         // if it is not repserver-management then respond with not understood. 

         ACL acl = conv.getACL( conv.getLatestMessageIndex() );

         if( acl.getOntology() == null || 
               !acl.getOntology().equalsIgnoreCase( RSCONSTANTS.REPSERVER_ONT ) ) 
         {
	    // sendNotUnderstood( acl );
	    done();
         }

         try 
         {
	    // Get content of message
	    String content = (String)acl.getContentObject();
            StringTokenizer st = new StringTokenizer(content); 

            String function = "";
	
            if ( st.hasMoreTokens() ) 
            {
               function = st.nextToken(); 
            } 
            else 
            {
               sendNotUnderstood(acl);  
               return;
            }

            // Decide which sub-tasks to spawn

            if ( function.equals( RSCONSTANTS.REGISTER) ) 
            {
	       newTask( new RegisterTask( acl), conv );
            } 
            else if ( function.equals( RSCONSTANTS.REPLICATE ) ) 
            {
	       newTask( new ReplicateTask( acl ), conv );
            } 
            else if ( function.equals( RSCONSTANTS.STATE ) ) 
            {
	       newTask( new StateTask( acl ), conv ); 
            } 
            else if ( function.equals( RSCONSTANTS.CREATE ) ) 
            {
	       newTask( new CreateTask( acl ), conv ); 
            } 
            else 
            {
               sendNotUnderstood( acl );
            }
         } 
         catch( Throwable t ) 
         {
            sendNotUnderstood( acl );
         } 
      } // handleRequest()
      

      public void doneRepServer_RegisterTask( Task t) 
      {
      }

    
      public void doneRepServer_ReplicateTask(Task t) 
      {
      }


      public void doneRepServer_StateTask(Task t) 
      {
      }

   
      public void doneRepServer_CreateTask(Task t)
      {
      }


   } // IdleTask()




   /** Register an agent
   */
   public class RegisterTask extends Task 
   {

      private ACL          _acl;
      private AgentID      _aid; 


      public RegisterTask(ACL msg) 
      {
         _acl = msg; 
      } // RegisterTask ()

   
      protected void startTask() 
      {
         // see which group this belongs to and which agent it is. 
         _aid = _acl.getSenderAID();
         StringTokenizer st = new StringTokenizer( (String)_acl.getContentObject() );  

         String rgname = "";

         if ( st.countTokens() != 2) 
         {
            sendNotUnderstood( _acl );
            done(); 
         } 
         else 
         {
            st.nextToken(); 
            rgname = st.nextToken();
          
            // see if this RepGroup exists on this server. 

            if ( !_repgroups.containsKey(rgname)) 
            { 
               writeMsg("Register Failed: Group " + rgname + " Does not exist");
               sendNotUnderstood( _acl );
               done();
            } 
            else 
            {
               ((RepGroup)_repgroups.get(rgname)).addAgent(new RepAgentID(_aid));

               // send an ACK (agree) 
               sendReply(_acl, FIPACONSTANTS.AGREE); 

               // send an Inform with the agentid of the proxy 
               sendReply(_acl, FIPACONSTANTS.INFORM, (Object) ((RepGroup)_repgroups.get(rgname)).getProxyAID());

               writeMsg("Register: Group " + rgname + " Agent: " + _aid.getName() + " Proxy: " + 
                     ((RepGroup)_repgroups.get(rgname)).getProxyAID().getName() + " Suceeded.");

               done();
            }
         } 
      } // startTask()

   } // RegisterTask



   /** One of the replicant groups active agents wants to send its 
       state. Respond and get the state. Then update the replicants 
       with this new state. 
   */
   public class StateTask extends Task 
   {

      private ACL          _acl;
      private AgentID      _aid; 

      private String       _group_name; 


      public StateTask(ACL msg) 
      {
         _acl = msg; 
      } // StateTask ()


      protected void startTask() 
      {
         // Get the content of the message and make sure it has 2 
         // tokens. 
         String content = (String) _acl.getContentObject();
         StringTokenizer st = new StringTokenizer(content);  

         if ( st.countTokens() != 2 ) 
         {
            sendNotUnderstood(_acl); 
            done();
            return; 
         }
 
         st.nextToken();
         _group_name = st.nextToken();

         if (_repgroups.containsKey(_group_name)) 
         {
            // if the named group exists, send back an Agree. 
            ACL resp = (ACL)_acl.clone();
            resp.setSenderAID( getAID() );
            resp.setReceiverAID( _acl.getSenderAID() );
            resp.setPerformative( FIPACONSTANTS.AGREE );
            forward( resp );
         } 
         else 
         { 
            // otherwise send a NotUnderstood
            sendNotUnderstood(_acl); 
            done();
         }
      } // startTask()
      

      /** The agent will send an INFORM with the state 
      */
      public void handleInform(Conversation conv) 
      {
         // the INFORM will hold the state information in its Content field. 
         ACL acl = conv.getACL( conv.getLatestMessageIndex() );
         byte[] state = (byte[])acl.getContentObject(); 
         // set the state in the RepGroup
         ((RepGroup)_repgroups.get(_group_name)).setState( state); 
         writeMsg("Received new state for replication group: " + _group_name); 
         removeConv(conv);

         // if this is a hot standby group start the update process
         if ( ((RepGroup)_repgroups.get(_group_name)).getReplicationType().equals(RSCONSTANTS.HOT_STANDBY) ) 
         {
System.out.println("UPdating repgroup members");
            newTask( new UpdateRepGroupTask( (RepGroup)_repgroups.get(_group_name))); 
         }
         else 
         {
            done();
         }

      } // handleInform()


      public void doneRepServer_UpdateRepGroupTask(Object o)
      {
         writeMsg("State update done for group: " + _group_name); 
         ((RepGroup)_repgroups.get(_group_name)).removeDead(); 
         done();
      } 


      public void timeoutRepServer_UpdateRepGroupTask(Task t)
      {
         writeMsg("State timeout done for group: " + _group_name); 
         ((RepGroup)_repgroups.get(_group_name)).removeDead(); 
         done();
      }

   } // StateTask 



   /** Create a replicant group
   **/ 
   public class ReplicateTask extends Task 
   {


      private ACL          _acl; 
      private String       _agent_name;
      private String       _class_name;
      private int          _N;
      private String       _agent_type;
      private int          _count; 
      private String       _replicationtype = RSCONSTANTS.HOT_STANDBY;


      public ReplicateTask( ACL acl) 
      {
         _acl = acl; 
      } // ReplicateTask


      protected void startTask() 
      {
         // Send Agree
         sendReply(_acl, FIPACONSTANTS.AGREE);

         String content = (String) _acl.getContentObject();

         StringTokenizer st = new StringTokenizer(content);  

         if ( st.countTokens() < 5 ) 
         {
            sendNotUnderstood(_acl); 
            done(); 
         }
         else 
         {
            st.nextToken();
            _agent_name = st.nextToken();
            _class_name = st.nextToken();
            _agent_type = st.nextToken();
            _N = Integer.parseInt(st.nextToken());
            if ( st.hasMoreTokens() ) 
            {
               _replicationtype = st.nextToken();
            }

            if (_repgroups.containsKey(_agent_name)) 
            {
               writeMsg("Replicant Group " + _agent_name + " Not created - Already there"); 
               done(); 
            } 
            else 
            {
               // create the rep group then create the agents. 
               RepGroup rg = new RepGroup(_platform_profile, _agent_name, _agent_type);
               rg.setReplicationType(_replicationtype);
               rg.getProxy().setLocalPostOffice( _localpo);
               _started_agents.put( rg.getProxy().getAID().getName(), rg.getProxy() );
               // add it to the list. 
               _repgroups.put(_agent_name, rg);
               writeMsg("Created Replicant Group " + _agent_name ); 

               _count = 0;
               newTask( new WaitTask(300));
            }
         }
      } // startTask()

      public void doneWaitTask(Task t)
      {
         if (_count == _N) 
         {
            sendReply(_acl, FIPACONSTANTS.INFORM);
            done();
         } 
         else 
         {
            startAgent( _agent_name + "." + _count, _class_name, "FIPA-OS", true, _agent_name);
            _count++;
            newTask( new WaitTask(300));
         }
      }


      /** Starts an agent
         @param agent_name Name of the Agent to start
         @param class_name Name of the class for this Agent
         @param owner Owner of the Agent
         @param async Start the Agent asyncronously?  
      */
      public void startAgent( final String agent_name, final String class_name,
            final String owner, boolean async, final String group_name ) 
      {
         // Check that no agents with the same name are currently alive
         if ( _started_agents.containsKey( agent_name ) ) 
         {
            writeMsg( "Agent \"" + agent_name + "\" is already running!" );
            return;
         }

         // Check if name is valid (contains no spaces!)
         if( agent_name.indexOf( " " ) != -1 ) 
         {
	    writeMsg( "Agent names cannot contain spaces!" );
	    return;
         }

         // A thread group so that the threads belonging to an agent can be kept track of
         ThreadGroup new_tg = new ThreadGroup( agent_name );

         // Object to lock on for sync. execution
         final boolean done[] = new boolean[1];

         // Start agent in new thread
         Thread t = new Thread( new_tg, new Runnable() 
         {
            public void run() 
            {
               try 
               {
                  // Add place-holder to list of active agents
                  // _started_agents_list.add( agent_name );
                  _started_agents.put( agent_name, new Object() );

                  // Setup expected parameters
                  Class args[] = new Class[] { String.class, String.class, String.class, AgentID.class, 
                        String.class, String.class };

                  Object params[] = new Object[] { _platform_profile, agent_name, owner, getAID(), 
                        RSCONSTANTS.REPLICANT, group_name };

                  DynamicClass dc = new DynamicClass( class_name, args, params );

                  // if the agent implements LocalMsg then set up the localpostoffice. 
                  try 
                  {
                     args = new Class[] {LocalPostOffice.class};
                     params = new Object[] {_localpo};
                     dc.invoke("setLocalPostOffice",args, params); 
                  }
                  catch ( Exception e)
                  {
                     // exception do nothing
                  }

                  // Add real FIPAOSAgent instance to map of started Agents
                  _started_agents.put( agent_name, dc.getObject() );
               } 
               catch (Throwable x) 
               {
                  writeMsg("Could Not Start agent " + _agent_name);
                  x.printStackTrace();
                  _started_agents.remove( agent_name );
               } 
               finally 
               {
                  synchronized( done ) 
                  {
                     done[0] = true;
	             done.notifyAll();
                  } // synchronized
	       } // finally
            } // run()
         } 
         );
		
         // Start thread async?
         if( !async ) 
         {
            // Wait for other thread to stop
            synchronized( done ) 
            {
	       t.start();
               while( !done[0] ) 
               {
                  try 
                  {
                     done.wait();
		  } 
                  catch( InterruptedException e ) 
                  {
		  }
	       } // while
	   } // synchronized
	} 
        else 
        {
	   // Just start other thread and continue
	   t.start();
        } // else 
      } //StartAgent()


   }

  
   /** Create N agents that will register with an existing replication group. 
   */
   public class CreateTask extends Task
   {
      private ACL          _acl; 
      private String       _agent_name;
      private String       _class_name;
      private int          _N;


      public CreateTask( ACL acl) 
      {
         _acl = acl; 
      } // 


      protected void startTask ()
      {
      }

   } // CreateTask



   /** Calls PingRepGroupTask for each RepGroup in this RepServer 
       When that completes it calls CheckRepGroup when that completes
       UpdateRepGroup.
   */
   public class PingRepGroupsTask extends Task 
   {

      private int _pc = 0;
      private int _cc = 0;


      public PingRepGroupsTask() 
      {
      }


      /** Start a PingRepGroup task for each RepGroup in this RepServer
      */
      protected void startTask() 
      {
 
         _pc  = _cc = _repgroups.size();

         String k = "";

         if ( _pc <= 0 ) 
         {
            // If _counter == 0, we're done! There are no groups.
            done();
         }
         else 
         {
            for (Enumeration e = _repgroups.keys(); e.hasMoreElements() ;) 
            {
               k = (String)e.nextElement();
               RepGroup rg = (RepGroup)_repgroups.get(k);
               RepAgentID raid = rg.getActive();

               // if there is no active, don't ping, proceed to check
               if ( raid != null )
                  newTask( new PingTask( rg, raid), RSCONSTANTS.PINGTIMEOUT ); 
               else 
                  decrementCountp(rg);
            } // for
         } // else

      } // startTask()


      /** the PingRepGroup task is done. 
      */
      public void doneRepServer_PingTask( Object o ) 
      {
         decrementCountp((RepGroup) o);
         // writeMsg("Ping Complete for group" + ((RepGroup)o).getName());
      }


      /** the PingRepGroup task has timed out. 
      */
      public void timeoutRepServer_PingTask( Task t ) 
      {
         if ( t instanceof PingTask ) 
         {
            PingTask p = (PingTask) t; 
            // set the agent's ping flag to false
            p.getGroup().setPing(p.getAgent(),false); 
            decrementCountp( p.getGroup() );
            writeMsg("Ping Timed-out for group" + p.getGroup().getName());
         }
      }


      /** the CheckRepGroupTask task is done. 
      */
      public void doneRepServer_CheckRepGroupTask(Object o) 
      {
         // writeMsg("Check Complete for group" + ((RepGroup)o).getName());
         decrementCountc((RepGroup) o);
      } 


      /** the CheckRepGroupTask task is has timed out. 
      */
      public void timeoutRepServer_CheckRepGroupTask(Task t) 
      {
         if ( t instanceof CheckRepGroupTask ) 
         {
            CheckRepGroupTask p = (CheckRepGroupTask) t; 
            decrementCountc( p.getGroup() );
            writeMsg("Check Timed-out for group" + p.getGroup().getName());
         }
      }


      /** For each PingTask that completes, decrement the count. And 
          spawn the CheckRepGroupTask 
      */
      private synchronized void decrementCountp(RepGroup rg) 
      {
         // Decrease counter
         _pc--;
         newTask( new CheckRepGroupTask(rg), RSCONSTANTS.PINGTIMEOUT); 

      } // decrementCount()


      /** For each Task that completes, decrement the count. 
      */
      private synchronized void decrementCountc(RepGroup rg) 
      {
         // Decrease counter
         _cc--;
         // If both are now at 0, we are done.
         if ( (_cc <= 0 ) && (_pc <= 0)) 
         {
            done();
         }

      } // decrementCount()


   } // PingRepGroupsTask 



   /** Sends a PING request to the given agent. Returns with a done(aid) call 
       where aid is the agent that was pinged. If the agent does not respond 
       this task will timeout. 
   */
   public class PingTask extends Task 
   {
      /** The agent to ping */
      private RepAgentID _agent_to_ping; 

      /** The group to ping */
      private RepGroup _group_to_ping; 

      /** Constructor. 
          @param aid set _agent_to_ping to this.  
      */
      public PingTask( RepGroup group, RepAgentID aid) 
      {
         _agent_to_ping = aid; 
         _group_to_ping = group;
      }

          
      /** Just returns the name of the Agent that this Task is pinging */
      public RepAgentID getAgent() 
      {
	 return _agent_to_ping;
      }

      /** Just returns the name of the Agent that this Task is pinging */
      public RepGroup getGroup() 
      {
	 return _group_to_ping;
      }

      /** Creates the ping message for the target Agent and sends it.
      */
      protected void startTask() 
      {
         // Create an ACLMessage and forward it to our target Agent
         ACL acl = new ACL();
         acl.setSenderAID( getAID() );
         acl.setReceiverAID( (AgentID)getAgent() );
         acl.setPerformative( FIPACONSTANTS.REQUEST );
         acl.setOntology( RSCONSTANTS.REPSERVER_ONT );
         acl.setProtocol( FIPACONSTANTS.FIPA_REQUEST);
         acl.setContentObject( RSCONSTANTS.PING );
         forward( acl );
      }


      /** When an agree comes back, call done().
      */
      public void handleAgree( Conversation conv ) 
      {
      }


      /** When an inform comes back, call done().
      */
      public void handleInform( Conversation conv ) 
      {
         done( getGroup() );
      }

   } // PingTask 



   /** Check a replicant group. Check that the active agent in the group 
       is still reachable. If it is, do nothing. If it is not, select a new 
       active agent.  
   */
   public class CheckRepGroupTask extends Task 
   {
      RepGroup _rg = null; 

      public CheckRepGroupTask(RepGroup rg) 
      {
         _rg = rg; 
      } // CheckRepGroupTask
 

      public RepGroup getGroup() 
      {
         return _rg;
      }

      /** Check all the repgroups that they are active. 
      */
      protected void startTask() 
      {
         boolean switched = false; 
         // find the active agent
         RepAgentID raid = _rg.getActive();   

         // if there is no active or the active was not pingable set a new one. 
         if ( (raid == null || !raid.getPing()) &&  _rg.setActive() ) 
         {
            raid = _rg.getActive();
            switched = true;
            newTask( new ActivateTask( _rg, raid), 10000 );
         } // if

         // remove dead agents from the _rg 
         _rg.removeDead(); 
         String reps = "";
         RepAgentID aids[] = _rg.getAgents();
         for (int i = 0; i < aids.length; i++) 
         {
            reps = reps + " " + aids[i].getName();
         }

         // writeMsg( "RepGroup Now is " + reps );

         if ( ! switched ) 
         {
            done(_rg);
         }

      } // startTask()


      /** When an agree comes back, we are done 
      */
      public void doneRepServer_ActivateTask(Task t)
      {
         done(_rg);
      } // handleAgree


      public void timeoutRepServer_ActivateTask(Task t)
      {
         done(_rg); 
      }

   } // CheckRepGroupTask


   public class ActivateTask extends Task
   {

      private RepGroup _rg;
      private RepAgentID _aid;

      private void sendRegister()
      {
         ACL reg = new ACL();
         reg.setPerformative( FIPACONSTANTS.REQUEST );
         reg.setSenderAID( getAID() );
         reg.setReceiverAID( (AgentID)_aid );
         reg.setContentObject( RSCONSTANTS.ACTIVATE );
         reg.setOntology( RSCONSTANTS.REPSERVER_ONT );
         reg.setProtocol( FIPACONSTANTS.FIPA_REQUEST );
         forward( reg );
      }

      public ActivateTask(RepGroup rg, RepAgentID aid)
      {
         _rg = rg;
         _aid = aid;
      }

      public void startTask()
      {
         if ( _rg.getReplicationType().equals(RSCONSTANTS.COLD_STANDBY)) 
            if ( _rg.getState() != null) 
               newTask( new UpdateAgentTask( _aid, _rg.getState()), RSCONSTANTS.PINGTIMEOUT ); 
            else 
               sendRegister();
         else 
            sendRegister();
      }

      public void handleAgree( Conversation c)
      {
	 writeMsg( "Switched Active Agent for Group " + _rg.getName() + " to: " + _rg.getActive().getName());
      }

      public void handleInform( Conversation c)
      {
         done();
      }

      /** A UpdateAgentTask completes normally. o is an RepAgentID 
          object pointing to the agent updated. 
      */
      public void doneRepServer_UpdateAgentTask(Object o) 
      {
         RepAgentID updated = (RepAgentID) o; 
         _rg.setUpdateTime((RepAgentID) o, new Date() );
         _rg.setPing((RepAgentID) o, true );
         writeMsg( "Updated Agent " + ((RepAgentID)o).getName());
         sendRegister();
      }


      /**
      */
      public void timeoutRepServer_UpdateAgentTask(Task t) 
      {
         if ( t instanceof PingTask ) 
         {
            PingTask p = (PingTask) t; 
            RepAgentID r = p.getAgent(); 
            _rg.setPing(r, false );
            writeMsg( "Update Failed and Ping not Set, Agent " + r.getName());
            done();
         }
      }

   }



   public class UpdateRepGroupTask extends Task 
   {

      private RepGroup _rg; 
      private int _count; 

      /** Constructor. 
          @param rg the replicant group to update.  
          _count is set to the number of replicants in the group. 
      */
      public UpdateRepGroupTask(RepGroup rg) 
      {
         _rg = rg;
         _count = _rg.countReps();
      }


      public RepGroup getGroup() 
      {
         return _rg;
      }


      /** For each agent in the replicant group, start a UpdateAgentTask.
      */
      protected void startTask() 
      {
         RepAgentID aids[] = _rg.getAgents();
         for (int i = 0; i < aids.length; i++) 
         { 
            if ( aids[i].getName().equals( _rg.getActive().getName() ))
            {
               // this is the active agent. decrement the count. 
               decrementCount();
            }
            else 
            {
               // start a task to update the agent
               newTask( new UpdateAgentTask( aids[i], _rg.getState()), RSCONSTANTS.PINGTIMEOUT ); 
            }
         }
         if ( _count <= 0 ) 
         {
            // If _counter == 0, we're done!
            done( _rg);
         }
      }


      /** A UpdateAgentTask completes normally. o is an RepAgentID 
          object pointing to the agent updated. 
      */
      public void doneRepServer_UpdateAgentTask(Object o) 
      {
         RepAgentID updated = (RepAgentID) o; 
         _rg.setUpdateTime((RepAgentID) o, new Date() );
         _rg.setPing((RepAgentID) o, true );
         writeMsg( "Updated Agent " + ((RepAgentID)o).getName());
         decrementCount(); 
      }


      /**
      */
      public void timeoutRepServer_UpdateAgentTask(Task t) 
      {
         if ( t instanceof PingTask ) 
         {
            PingTask p = (PingTask) t; 
            RepAgentID r = p.getAgent(); 
            _rg.setPing(r, false );
            writeMsg( "Update Failed and Ping not Set, Agent " + r.getName());
            decrementCount(); 
         }
      }


      /**
      */
      public void errorRepServer_UpdateAgentTask(Throwable th, Task t) 
      {
         if ( t instanceof UpdateAgentTask ) 
         {
            UpdateAgentTask p = (UpdateAgentTask) t; 
            RepAgentID r = p.getAgent(); 
            _rg.setPing(r, false );
            writeMsg( "Update Failed and Ping not Set, Agent " + r.getName());
            decrementCount(); 
         }
      }


      /** For each UpdateAgentTask that completes, decrement the count. If 
          count goes to 0, this task is finished, call done().
      */
      private synchronized void decrementCount() 
      {
         _count--;

         // Check if we've reached 0
         if ( _count <= 0 ) 
         {
            // If _counter == 0, we're done!
            done(_rg);
         }

      } // decrementCount()

   } // UpdateRepGroupTask


   /** Sends a LOADSTATE request to the given agent. Returns with a done(aid) call 
       where aid is the agent that was updated. If the agent does not respond 
       this task will timeout. 
   */
   public class UpdateAgentTask extends Task {

      /** The agent to update */
      private RepAgentID _agent_to_ping; 
      
      private byte[] _state;
      
      private boolean _statesent;

      /** Constructor. 
          @param aid set _agent_to_ping to this.  
      */
      public UpdateAgentTask( RepAgentID aid, byte[] s) 
      {
         _agent_to_ping = aid; 
         _state = s;
         _statesent = false;
      }

          
      /** Just returns the name of the Agent that this Task is pinging */
      public RepAgentID getAgent() 
      {
	 return _agent_to_ping;
      }


      /** Creates the LOADSTATE message for the target Agent and sends it.
      */
      protected void startTask() 
      {
         // Create an ACLMessage and forward it to our target Agent
         ACL acl = new ACL();
         acl.setPerformative( FIPACONSTANTS.REQUEST );
         acl.setSenderAID( getAID() );
         acl.setReceiverAID( (AgentID)getAgent() );
         acl.setContentObject( RSCONSTANTS.LOADSTATE );
         acl.setOntology( RSCONSTANTS.REPSERVER_ONT );
         forward( acl );
      }


      /** When an agree comes back, send the state. 
      */
      public void handleAgree( Conversation conv ) 
      {
         // send the state
         _statesent = true;
         ACL acl = conv.getACL( conv.getLatestMessageIndex() );
         ACL resp = (ACL)acl.clone();  
         resp.setPerformative( FIPACONSTANTS.INFORM );
         resp.setSenderAID( getAID() );
         resp.setReceiverAID( (AgentID)getAgent() );
         resp.setContentObject( _state );
         resp.setOntology( RSCONSTANTS.REPSERVER_ONT );
         forward( resp );
         removeConv( conv);
	 done( getAgent() );
      }



   } // UpdateAgentTask 


   /** Test-harness
   */
   public static void main( String args[] ) 
   {
      if ( args == null || args.length < 2 ) 
      {
         System.out.println( "Usage: java fipaos.repserver.RepServer <platform.profile> <name> <gui|nogui>" );
         return;
      }
      new RepServer( args[0], args[1], "FIPA-OS", args[2] );

   } // main()

} // ReqAgent
