/* 
   Test Agent Server - CountAgent V1.0, November 12, 2001

   Written by: 
      Alan Fedoruk                                           -- amf673@mail.usask.ca
      MADMUC Lab                                             -- http://bistrica.usask.ca/madmuc/index.htm
      Dept. of Computer Science, University of Saskatchewan. -- http://www.cs.usask.ca

   Copyright (c) 2001 Alan M. Fedoruk
*/


package testAS; 

// Import the fipa-os classes
import fipaos.agent.*;
import fipaos.agent.conversation.*;
import fipaos.agent.task.*;
import fipaos.mts.*;
import fipaos.ont.fipa.*;
import fipaos.ont.fipa.fipaman.*;
import fipaos.parser.acl.ACLMessage;

// Import registration exception classes
import fipaos.platform.ams.AMSRegistrationException;
import fipaos.platform.df.DFRegistrationException;
import fipaos.util.DIAGNOSTICS;

// Import java Utilities
import java.util.*;
import java.io.*;

import agentserver.ASCONSTANTS; 


/** CountAgent is a demo agent that increments a counter and prints the 
    values on a simple GUI. 
    CountAgent Will respond to a REQUEST by sending back an INFORM 
    with the current counter value. 
*/
public class CountAgent extends FIPAOSAgent {

   /** Counter */
   private int _N = 0;
  
   /** This Agent type */
   private static final String AGENT_TYPE = "count-agent";

   /**  GUI for this CountAgent */
   private CountAgentGUI _gui;


   /** Agent constructor - sets up an IdleTask as the default
       ListenerTask for incoming messages, and starts off registration
       with the AMS and DF
   */
   public CountAgent( String platform, String name, String ownership ) {
   
      super( platform, name, ownership );
    
      _N = 0; 
 
      // Create a GUI for this Agent so we can see what is going on.
      try 
      {
         _gui = new CountAgentGUI( getAID().getName(), this );
      } 
      catch ( Throwable t ) 
      { 
      }
    
      // Instantiate a new IdleTask to receieve all incoming 
      // communications which are not associated with any-other Task

      setListenerTask( new IdleTask() );
  
      // Now we can attempt to register with the local AMS (i.e. the platform)

      try 
      {
         registerWithAMS();
         DIAGNOSTICS.println( "Registered with AMS", this, DIAGNOSTICS.LEVEL_MAX );
         _gui.write( "Registered with AMS" );
         _gui.write( "Agent Name = " + getAID());

      } 
      catch (AMSRegistrationException amsre) 
      {
         // registration failed for some reason. Display the exception
         DIAGNOSTICS.println(amsre, this, DIAGNOSTICS.LEVEL_MAX );
      
         String reason = amsre.getExceptionReason();
      
         // If the exception reason is not "already-registered", we can 
         // probably ignore the failure
         // (since this Agent was probably not de-registered the last time 
         // it ran!)

         if ( reason == null || !reason.equals( FIPAMANCONSTANTS.AGENT_ALREADY_REGISTERED)) 
         {
            // Shutdown nicely :)
            shutdown();
            return;
         }
      
      } // catch 
    
      // Now we can register with the local DF
      try 
      {
         registerWithDF( AGENT_TYPE );
         DIAGNOSTICS.println("Registered with DF", this, DIAGNOSTICS.LEVEL_MAX );
         _gui.write("Registered with DF" );

      } 
      catch (DFRegistrationException dfre) 
      {
         // the DF registration failed for some reason
         // Display the exception
         DIAGNOSTICS.println(dfre, this, DIAGNOSTICS.LEVEL_MAX );
      
         String reason = dfre.getExceptionReason();
      
         if ( reason == null || !reason.equals(
               FIPAMANCONSTANTS.AGENT_ALREADY_REGISTERED) ) 
         {
            shutdown();
            return;
         }
      
      } // end catch

      _tm.newTask( new CountTask());
      
   } // CountAgent()
  



   /** Invoked when this Agent is about to be shutdown - this overrides the
       shutdown() method in the FIPAOSAgent class to enable the CountAgentGUI
       associated with this CountAgent to be removed
   */
   public synchronized void shutdown() 
   {
      if (registeredWithDF() == true) 
      {
         // Attempt to deregister with DF
         try 
         {
            deregisterWithDF();
            DIAGNOSTICS.println("Deregistered with DF", this, DIAGNOSTICS.LEVEL_MAX );
         } 
         catch (DFRegistrationException dfre) 
         {
            // A problem deregistering occured..... we can obtain 
            // the reason though!
            String reason = dfre.getExceptionReason();

            DIAGNOSTICS.println(dfre + ":" + reason, this, DIAGNOSTICS.LEVEL_MAX ); 
         } // end catch
      } // if 
    
      if (registeredWithAMS() == true) 
      {
         // Attempt to deregister with AMS
         try 
         {
            deregisterWithAMS();
            DIAGNOSTICS.println("Deregistered with AMS", this, DIAGNOSTICS.LEVEL_MAX ); 
         } 
         catch (AMSRegistrationException amsre) 
         {
            // A problem deregistering occured..... we can obtain the 
            // reason though!
            String reason = amsre.getExceptionReason();
            DIAGNOSTICS.println(amsre + ":" + reason, this, DIAGNOSTICS.LEVEL_MAX );
         } // catch
      } // if 
     
      // Dispose of the GUI
      if( _gui != null ) 
      {
         _gui.dispose();
      }
    
      // Now call shutdown in the agent shell to release the core components
      super.shutdown();

   } // shutdown()
  

   /** This method is invoked when GUI should be shown - generally
       this is invoked by the AgentLoader when the Agent is in the List
       of running Agents - the default implementation in FIPAOSAgent does 
       nothing.
   */
   public void activate() 
   {
      // Check if we've got a GUI, if so, show it
      if( _gui != null ) 
      {
         _gui.setVisible( true );
      }
   } // activate()
    

   /** return a bytestream that represents this agents state 
   */
   public byte[] getAState() 
   {
      try 
      {
         ByteArrayOutputStream bos = new ByteArrayOutputStream();
         ObjectOutputStream oos = new ObjectOutputStream(bos);
         // DIAGNOSTICS.println("Get Agent State " + _N, DIAGNOSTICS.LEVEL_4);

         oos.writeInt(_N);
   
         bos.close();
         oos.close();
         return ( bos.toByteArray());
      } 
      catch ( Exception e) 
      {
         DIAGNOSTICS.println("Failed to Get Agent State", DIAGNOSTICS.LEVEL_4);
         return ( new byte[1]);
      }

   } // getAState 


   /** accept a bytestream that represents this agents state 
       and apply it to the agent
   */
   public boolean setAState(byte[] b) 
   {
      try 
      {
         ByteArrayInputStream bis = new ByteArrayInputStream(b);
         ObjectInputStream ois = new ObjectInputStream(bis);

         _N = ois.readInt();

         // DIAGNOSTICS.println("Set Agent State to" + _N, DIAGNOSTICS.LEVEL_4);

         ois.close();
         bis.close();
      } 
      catch ( Exception e) 
      {
         DIAGNOSTICS.println("Failed to Set Agent State", DIAGNOSTICS.LEVEL_4);
         e.printStackTrace();
         return false;
      }
      return true;

   } // setAState

   /**
    * Send a reply message in response to the given ACL message
    *
    * @param acl The ACL object to reply to
    * @param reason Any additional message
    **/
   public boolean sendReply( ACL orig_acl, String performative, String content )
   {
      boolean done = false;
      
      try
      {
         // Duplicate ACL
         ACL acl = (ACL) orig_acl.clone();
            
         // Setup message
         acl.setPerformative( performative );
         acl.setReceiverAID( acl.getSenderAID() );
         acl.setSenderAID( getAID() );
         acl.setInReplyTo( acl.getReplyWith() );
         acl.setReplyWith( null );
            
         // Set content
         acl.setContentObject( content );
            
         // Attempt to send the agree
         forward( acl );
            
         // We're done
         done = true;
      }
      catch( Throwable t )
      {
         DIAGNOSTICS.println( "Error occured when sending agree", t, this,
         DIAGNOSTICS.LEVEL_MAX );
      }
      
      return done;
   } // sendReply()
   

   /** IdleTask, listen for incoming messages 
   */
   public class IdleTask extends Task 
   {

      public IdleTask() 
      {
      } 
    
    
      protected void startTask() 
      {
      } 

      public void handleRequest( Conversation conv ) 
      {
         // Get the message
         ACL acl = conv.getACL( conv.getLatestMessageIndex() );

         newTask( new HandleRequestTask(acl), conv);

      } // handleRequest()


      public void doneWaitTask(Task t) 
      {
      }

      public void doneCountAgent_HandleRequestTask( Task t) 
      {
      }
    
   } // IdleTask()


   /** CountTask every 5 seconds increments the agents counter 
   */
   public class CountTask extends Task 
   {

      public CountTask() 
      {
      }


      protected void startTask() 
      {
         newTask( new WaitTask( 5000 ) );
      }


      public void doneWaitTask( Task t ) 
      {
         newTask( new WaitTask( 5000 ) );
         _gui.write("Count:" + (_N++));
      } // doneWaitTask()
   }


   /** handle a request coming to this agent. Requests
       are handled by responding with an AGREE/INFORM
       sequence with the current counter value. 
   */
   public class HandleRequestTask extends Task
   {

      /* the incoming message */
      private ACL _acl; 


      public HandleRequestTask( ACL acl)
      {
         _acl = acl;
      }


      protected void startTask() 
      {
         // construct the agree
         sendReply(_acl, FIPACONSTANTS.AGREE, ""); 

         sendReply(_acl, FIPACONSTANTS.INFORM, (String) " "+_N);

         done();

      } // startTask()

   }

   /**
   * Test-harness
   */
   public static void main( String args[] ) 
   {
      if ( args == null || args.length < 2 ) 
      {
         System.out.println( "Usage: java fipaos.tutorial.CountAgent <platform.profile> <name>" );
         return;
      }
      try 
      {
         new CountAgent( args[0], args[1], "FIPA-OS"); 
      } catch ( Exception e ) 
      {
          System.out.println(e);
      }

   } // main()

} // CountAgent
